"""Nested model relationships for Aquarium models.
"""

import inflection

from pydent.base import ModelBase
from pydent.marshaller import fields
from pydent.marshaller.exceptions import MarshallerRelationshipError
from pydent.utils import MagicList


class One(fields.Relation):
    """
    Defines a single relationship with another model.
    Subclass of :class:`pydent.marshaller.Relation`.
    """

    def __init__(self, model, *args, callback=None, params=None, **kwargs):
        """
        One initializer. Uses "find" callback by default.

        :param model: target model
        :type model: basestring
        :param args: other args for fields.Nested relationship
        :type args: ...
        :param attr: attribute to use to find model
        :type attr: basestring
        :param kwargs: other kwargs for fields.Nested relationship
        :type kwargs: ...
        """
        if callback is None:
            callback = ModelBase.find_callback.__name__
        super().__init__(model, *args, callback=callback, params=params, **kwargs)


class Many(fields.Relation):
    """
    Defines a many relationship with another model.
    Subclass of :class:`pydent.marshaller.Relation`.
    """

    def __init__(self, model, *args, callback=None, params=None, **kwargs):
        """
        Many initializer. Uses "where" callback by default.

        :param model: target model
        :type model: basestring
        :param args: other args for fields.Nested relationship
        :type args: ...
        :param attr: attribute to use to find model
        :type attr: basestring
        :param kwargs: other kwargs for fields.Nested relationship
        :type kwargs: ...
        """
        if callback is None:
            callback = ModelBase.where_callback.__name__
        super().__init__(model, *args, default=MagicList,
                         many=True, callback=callback, params=params, **kwargs)


class HasMixin:
    """
    Mixin for adding the 'set_ref' method. 'set_ref' builds a 'ref' and 'attr' attributes
    """

    def set_ref(self, model=None, ref=None, attr=None):
        """
        Sets the 'ref' and 'attr' attributes. These attributes are used to
        defined parameters for
        :class:`pydent.marshaller.Relation` classes.

        For example:

        .. code-block:: python

            relation # HasOne, HasMany, or HasManyGeneric, etc.
            relation.set_ref(ref="parent_id")
            relation.ref    # "parent_id"
            relation.attr   # "id"

            relation.set_ref(model="SampleType")
            relation.ref   # "sample_type_id"
            relation.attr  # "id"

            relation.set_ref(attr="name", model="OperationType")
            relation.ref   # "operation_type_name
            relation.attr  # "name"
        """
        self.ref = ref
        self.attr = attr

        if not self.attr:
            self.attr = 'id'
        if ref:
            self.ref = ref
        else:
            if not self.attr:
                raise MarshallerRelationshipError(
                    "'attr' is None. Relationship '{}' needs an 'attr' and 'model' parameters".format(self.__class__))
            if not model:
                raise MarshallerRelationshipError(
                    "'model' is None. Relationship '{}' needs an 'attr' and 'model' parameters".format(self.__class__))
            self.ref = "{}_{}".format(inflection.underscore(model), self.attr)


class HasOne(HasMixin, One):
    def __init__(self, model, attr=None, ref=None, **kwargs):
        """
        HasOne initializer. Uses the "get_one_generic" callback and automatically
        assigns attribute as in the following:

        .. code-block:: python

            model="SampleType", attr="id" # equiv. to 'lambda self: self.sample_type_id.'

        :param model: model name of the target model
        :type model: basestring
        :param attr: attribute to append underscored model name
        :type attr: basestring
        """
        self.set_ref(model=model, attr=attr, ref=ref)
        super().__init__(model, params=(lambda slf: getattr(slf, self.ref)), **kwargs)

    def __repr__(self):
        return "<HasOne (model={}, params=lambda self: self.{})>".format(self.model, self.ref)


class HasManyThrough(HasMixin, Many):
    """
    A relationship using an intermediate association model.
    Establishes a Many-to-Many relationship with another model
    """

    def __init__(self, model, through, attr="id", ref=None, **kwargs):
        self.set_ref(model=model, attr=attr, ref=ref)

        # e.g. PlanAssociation >> plan_associations
        through_model_attr = inflection.pluralize(
            inflection.underscore(through))

        # e.g. {"id": x.operation_id for x in self.plan_associations
        def params(slf):
            through_model = getattr(slf, through_model_attr)
            if through_model is None:
                return None
            return {attr: [getattr(x, self.ref) for x in getattr(slf, through_model_attr)]}
        super().__init__(model, params=params, **kwargs)


class HasMany(HasMixin, Many):
    """
    A relationship that establishes a One-to-Many relationship with another model.
    """

    def __init__(self, model, ref_model=None, attr=None, ref=None, **kwargs):
        """
        HasMany relationship initializer

        :param model: Model class name for this relationship
        :type model: str
        :param ref_model: Reference model name of the model owning this relationships.

        .. code-block:: python

            @add_schema
            class Author(ModelBase):
                fields=dict(books=HasMany("Book", "Author"))  # search for books using 'author_id'

        :type ref_model: str
        :param attr: Attribute name to use with reference model (default='id').
                     For example "Author" => 'author_id'
        :type attr: str
        :param ref: The reference to use to find models. If none, a reference is
                    built from the 'ref_model' and 'attr' parameters
        :type ref: str
        """
        if ref_model is None and ref is None:
            raise MarshallerRelationshipError("'{}' needs a 'ref_model' or 'ref' parameters to initialize".format(
                self.__class__.__name__))
        self.set_ref(model=ref_model, attr=attr, ref=ref)

        def params(slf): return {self.ref: getattr(slf, self.attr)}
        super().__init__(model, params=params, **kwargs)


class HasManyGeneric(HasMany):
    """
    Establishes a One-to-Many relationship using 'parent_id' as the attribute
    to find other models.
    """

    def __init__(self, model, **kwargs):
        super().__init__(model, ref="parent_id", attr="id", **kwargs)



# TRIDENT

__title__ = 'Trident'
__description__ = 'Aquarium scripting'
__url__ = 'https://klavinslab.org/trident'
__version__ = '0.0.22'
__author__ = 'Justin Vrana, Eric Klavins, Ben Keller'
__author_email__ = "jvrana@uw.edu"



"""
Session interfaces for interacting with Aquarium.

Session interfaces are created by an AqSession instance and use an AqHTTP
instance to make http requests to Aquarium.

Generally, Trident users should be unable to directly access AqHTTP method to
prevent potentially damaging requests or revealing sensitive information.
Session interfaces define a particular way an AqHTTP is used.

Interfaces are roughly grouped into:

(1) ModelInterface - for finding and retrieving Aquarium models and information
(2) CreateInterface - for creating new objects (e.g. new Samples) in Aquarium
(3) UpdateInterface - for updating objects in Aquarium

Example:
    session1 = AqSession(login, password, aquarium_url)
    session1.User
    # => a ModelInterface instance

    session1.User.find(1)
    # use model interface with User to find User(id=1)

    session1.aqhttp.post(...)
    # not allowed

    session1.create
    # => a CreateInterface

    session1.create.samples(list_of_samples)
    # creates samples from a list by calling method CreateInterface.samples
"""

from inflection import pluralize, underscore
from .base import ModelRegistry
from .exceptions import TridentRequestError
from .utils import url_build


class SessionInterface(object):
    """
    Generic session interface.

    Trident users should only be able to make requests through a
    SessionInterface to avoid making arbitrary and potentially damaging
    http requests.
    """

    def __init__(self, aqhttp, session):
        """
        Initializer for SessionInterface

        :param aqhttp: aqhttp instance for this interface
        :type aqhttp: AqHTTP
        :param session: session instance for this interface
        :type session: AqSession
        """
        self.aqhttp = aqhttp
        self.session = session


class UtilityInterface(SessionInterface):
    """
    Miscellaneous requests for creating, updating, etc.
    """
    # TODO: have ability to save new properties
    def create_samples(self, samples):
        json = [s.dump() for s in samples]
        return self.aqhttp.post('browser/create_samples', {"samples": json})

    def create_items(self, items):
        return [self.aqhttp.get('items/make/{}/{}'.format(
                i.sample.id, i.object_type.id))
                for i in items]

    def create_sample_type(self, sample_type):
        """
        Creates a new sample type
        """
        st_data = sample_type.dump(relations=('field_types'))
        result = self.aqhttp.post('sample_types.json', json_data=st_data)
        sample_type.reload(result)
        return sample_type

    def create_object_type(self, object_type):
        """
        Creates a new object type
        """
        ot_data = object_type.dump()
        result = self.aqhttp.post('object_types.json', json_data=ot_data)
        object_type.reload(result)
        return object_type

    def create_operation_type(self, operation_type):
        """
        Creates a new operation type
        """
        op_data = operation_type.dump(include={
            'field_types': {
                'allowable_field_types': {}
            },
            'protocol': {},
            'cost_model': {},
            'documentation': {},
            'precondition': {}
        })

        # Format 'sample_type' and 'object_type' keys for afts
        for ft_d, ft in zip(
                op_data['field_types'], operation_type.field_types):
            for aft_d, aft in zip(
                    ft_d['allowable_field_types'], ft.allowable_field_types):
                aft_d['sample_type'] = {'name': aft.sample_type.name}
                aft_d['object_type'] = {'name': aft.object_type.name}

        result = self.aqhttp.post('operation_types.json', json_data=op_data)
        operation_type.reload(result)
        return operation_type

    def estimate_plan_cost(self, plan):
        """
        Estimates the plan cost
        """
        result = self.aqhttp.post('launcher/estimate', {'id': plan.id})
        total = 0
        for operation in plan.operations:
            for cost in result['costs']:
                if cost['id'] == operation.id:
                    operation.cost = cost
                    total += (cost['labor'] * cost['labor_rate'] +
                              cost['materials']) * cost['markup_rate']
        return total

    def create_plan(self, plan):
        """
        Saves a plan
        """
        pre_ops = plan.operations
        plan_data = plan.to_save_json()
        result = self.aqhttp.post(
            "plans.json?user_id={}".format(self.session.current_user.id),
            json_data=plan_data)
        plan.reload(result)
        post_ops = plan.operations
        for i in range(len(pre_ops)):
            pre_ops[i].reload(post_ops[i])
        return plan

    def save_plan(self, plan):
        plan_data = plan.to_save_json()
        result = self.aqhttp.put(
            "plans/{}.json?user_id={}".format(plan.id,
                                              self.session.current_user.id),
            json_data=plan_data
        )
        plan.reload(result)
        return plan

    def delete_plan(self, plan):
        self.aqhttp.delete('plans/{}'.format(plan.id))

    def replan(self, plan_id):
        """
        Copies a plan
        """
        result = self.aqhttp.get(
            'plans/replan/{plan_id}'.format(plan_id=plan_id))
        plan_copy = self.session.Plan.load(result)
        return plan_copy

    def start_job(self, job_id):
        result = self.aqhttp.get(
            'krill/start?job={job_id}'.format(job_id=job_id))
        return result

    def batch_operations(self, operation_ids):
        """
        Batches operations from a list of operation_ids
        """
        self.aqhttp.post('operations/batch', json_data=operation_ids)

    def unbatch_operations(self, operation_ids):
        """
        Unbatches operations from a list of operation_ids
        """
        self.aqhttp.post('operations/batch', json_data=operation_ids)

    def set_operation_status(self, operation_id, status):
        """
        Sets an operation's status
        """
        msg = 'operations/{oid}/status/{status}'.format(
            oid=operation_id,
            status=status)
        self.aqhttp.get(msg)

    def job_debug(self, job_id):
        """
        Runs debug on a job with id=job_id
        """
        self.aqhttp.get('krill/debug/{jid}'.format(jid=job_id))

    def submit_plan(self, plan, user, budget):
        """
        Submits a plan
        """
        user_query = "&user_id=" + str(user.id)
        budget_query = "?budget_id=" + str(budget.id)
        result = self.aqhttp.get('plans/start/' + str(plan.id) +
                                 budget_query + user_query)
        return result

    def update_code(self, code):
        """
        Updates code for a operation_type
        """
        controller = underscore(
            pluralize(code.parent_class))

        code_data = {
            "id": code.parent_id,
            "name": code.name,
            "content": code.content
        }
        result = self.aqhttp.post(url_build(controller, "code"), code_data)
        if "id" in result:
            code.id = result["id"]
            code.parent_id = result["parent_id"]
            code.updated_at = result["updated_at"]
        else:
            raise TridentRequestError(
                "Unable to update code object {}".format(code_data), result)

    def compatible_items(self, sample_id, object_type_id):
        """
        Find items compatible with the field value.
        """
        result = self.aqhttp.post("json/items", {
            "sid": sample_id,
            "oid": object_type_id
        })
        items = []
        for element in result:
            print(element)
        return items


class ModelInterface(SessionInterface):
    """
    Makes requests using AqHTTP that are model specific.
    Establishes a connection between a session object and an Aquarium model.
    """

    def __init__(self, model_name, aqhttp, session):
        super().__init__(aqhttp, session)
        self.model = ModelRegistry.get_model(model_name)

    @property
    def model_name(self):
        """
        Alias for self.model.__name__
        """
        return self.model.__name__

    def _post_json(self, data, get_from_history_ok=False):
        """
        Posts a json request to session for this interface.
        Attaches raw json and this session instance to the models it retrieves.
        """
        data_dict = {'model': self.model_name}
        data_dict.update(data)

        post_response = None
        try:
            post_response = self.aqhttp.post(
                'json',
                json_data=data_dict,
                get_from_history_ok=get_from_history_ok)
        except TridentRequestError as err:
            if err.response.status_code == 422:
                return None
            raise err

        if post_response:
            return self.load(post_response)

    def load(self, post_response):
        """
        Loads model instance(s) from data.
        Model instances will be of class defined by self.model.
        If data is a list, will return a list of model instances.
        """
        many = isinstance(post_response, list)
        models = self.model.load(post_response, many=many)
        if many:
            assert len(post_response) == len(models)
            for model in models:
                model.connect_to_session(self.session)
        else:
            models.connect_to_session(self.session)
        return models

    def get(self, path):
        """
        Makes a generic get request
        """
        response = self.aqhttp.get(path)
        return self.load(response)

    def find(self, model_id):
        """
        Finds model by id
        """
        return self._post_json({"id": model_id}, get_from_history_ok=True)

    def find_by_name(self, name):
        """
        Finds model by name
        """
        return self._post_json({"method": "find_by_name", "arguments": [name]},
                               get_from_history_ok=True)

    def array_query(self, method, args, rest, opts=None):
        """
        Finds models based on a query
        """
        if opts is None:
            opts = {}
        options = {"offset": -1, "limit": -1, "reverse": False}
        options.update(opts)
        query = {"model": self.model.__name__,
                 "method": method,
                 "arguments": args,
                 "options": options}
        query.update(rest)
        res = self._post_json(query)  # type: dict
        return res

    def all(self, rest=None, opts=None):
        """
        Finds all models
        """
        if rest is None:
            rest = {}
        if opts is None:
            opts = {}
        options = {"offset": -1, "limit": -1, "reverse": False}
        options.update(opts)
        return self.array_query("all", [], rest, options)

    def where(self, criteria, methods=None, opts=None):
        """
        Finds models based on criteria
        """
        if methods is None:
            methods = {}
        if opts is None:
            opts = {}
        options = {"offset": -1, "limit": -1, "reverse": False}
        options.update(opts)
        return self.array_query("where", criteria, methods, options)

    def patch(self, model_id, json_data):
        """
        Makes a patch or update from json_data
        """
        models_name = pluralize(underscore(self.model_name))
        result = self.aqhttp.put('{models}/{model_id}'.format(
            model_id=model_id, models=models_name), json_data=json_data)
        return result

    def new(self, *args, **kwargs):
        """
        Creates a new model instance
        """
        instance = object.__new__(self.model)
        instance._session = None
        instance.connect_to_session(self.session)
        instance.__init__(*args, **kwargs)
        return instance



import json
import warnings
from copy import deepcopy

from pydent.marshaller.exceptions import (MarshallerCallbackNotFoundError,
                                          MarshallerRelationshipError)
from pydent.marshaller.schema import MODEL_SCHEMA
from pydent.utils import pformat


class MarshallerBase(object):
    """
    Base class for marshalling and unmarshalling. Used in conjunction with
    :func:`pydent.marshaller.schema.add_schema`

    Example Usage:


    .. code-block:: python

        from pydent.marshaller import MarshallerBase, add_schema

        @add_schema
        class UserModel(MarshallerBase):
            fields = dict(
                ignore=("password",),
                budgets=Relation("Budget", callback="find_budgets",
                                 params=lambda self: self.id)
            )

            def find_budgets(user_id):
                # find budgets with user_id
                pass

        # load from JSON
        u = UserModel.load(user_data)

        # dump to JSON
        u.dump()
    """

    model_schema = None
    save_attr = True

    @classmethod
    def create_from_json(cls, *args, data=None, **kwargs):
        """
        Special constructor method for instantiating with JSON data

        :param data: json data to update instance attributes
        :type data: dict
        :return: model instance
        :rtype: MarshallerBase (or subclass)
        """
        inst = cls(*args, **kwargs)
        if data is None:
            data = {}
        vars(inst).update(data)
        return inst

    @classmethod
    def get_schema_class(cls):
        """Return the Schema class associated with this model"""
        return getattr(cls, MODEL_SCHEMA)

    @classmethod
    def create_schema_instance(cls, *schema_args, **schema_kwargs):
        """Create a Schema instance for loading or dumping"""
        schema_class = cls.get_schema_class()
        if schema_class:
            schema = cls.get_schema_class()(*schema_args, **schema_kwargs)
            schema.context = schema_kwargs
            return schema

    @classmethod
    def get_relationships(cls):
        """
        Get the name: relationship
        (:class:`pydent.marshaller.relation.Relation`)
        dictionary
        """
        schema_class = cls.get_schema_class()
        if schema_class:
            return cls.get_schema_class().relationships
        return {}

    @property
    def relationships(self):
        """
        Get the name: relationship
        (:class:`pydent.marshaller.relation.Relation`)
        dictionary.
        """
        return self.__class__.get_relationships()

    @classmethod
    def load(cls, data, *args, only=None, exclude=(), all_relations=False,
             prefix='', strict=None, many=False, relations=(), load_only=(),
             dump_only=(), partial=False, **kwargs):
        """Loads the model instance from JSON. For parameter options, refer to
        :class:`pydent.marshaller.schema.DynamicSchema`"""

        schema_kwargs = dict(
            only=only,
            exclude=exclude,
            relations=relations,
            all_relations=all_relations,
            prefix=prefix,
            strict=strict,
            many=many,
            load_only=load_only,
            dump_only=dump_only,
            partial=partial,
        )
        schema_kwargs.update(kwargs)

        schema = cls.create_schema_instance(*args, **schema_kwargs)
        if schema:
            return schema.load(data).data

        # TODO: violates an invariant, throw an exception
        warnings.warn(
            "Cannot dump! No schema attached to '{}'".format(cls.__name__))

    def dump(self, only=None, include=(), exclude=(), relations=(),
             all_relations=False, prefix='', strict=None,
             many=False, load_only=(), dump_only=(),
             partial=False, **kwargs):
        """Dumps the model instance to a String. For parameter options, refer
        to :class:`pydent.marshaller.schema.DynamicSchema`

        :param tuple only: A list or tuple of fields to serialize. If `None`,
                all non relationship fields will be serialized.
        :param str|list|tuple|set|dict include: Nested relationships to include
                in the serialization. This accepts a string (for one nested
                relationship), a tuple (for multiple relationships) or a dict
                (for recursively applied serialization). A dictionary of dump
                options 'opts' can be recursively applied to each serialization.
        :param tuple exclude: A list or tuple of fields to exclude from the
                serialized result. Nested fields can be represented with dot
                delimiters.
        :param tuple relations: A list or tuple of fields to include during
                serialization. Unlike 'include', this will not be applied
                recursively
        :param boolean all_relations: whether to dump all relationship during
                deserialization
        :param str prefix: Optional prefix that will be prepended to all the
                serialized field names.
        :param bool strict: If `True`, raise errors if invalid data are passed
                in instead of failing silently and storing the errors.
        :param bool many: Should be set to `True` if ``obj`` is a collection so
                that the object will be serialized to a list.
        :param dict context: Optional context passed to :class:`fields.Method`
                and :class:`fields.Function` fields.
        :param tuple load_only: A list or tuple of fields to skip during
                serialization
        :param tuple dump_only: A list or tuple of fields to skip during
                deserialization, read-only fields
        :param bool|tuple partial: Whether to ignore missing fields. If its
                value is an iterable, only missing fields listed in that
                iterable will be ignored.


        Example Usage using 'only' to dump only particular relationships

        .. code-block:: python

            # dump 'name' and nested relationship 'sample_type'
            mydump = mymodel.dump(only=('name', 'sample_type'))

        Example Usage using 'include' to dump nested relationships:

        .. code-block:: python

            # dump nested relationships 'sample_type' and 'samples' in 'sample_type'
            # for 'sample_type' dump only 'name' in addition to 'samples'
            mydump = mymodel.dump(include={'sample_type': { 'samples': {}, 'opts': {'only': ['name']} } })

        """
        schema_kwargs = dict(
            only=only,
            exclude=exclude,
            relations=relations,
            all_relations=all_relations,
            prefix=prefix,
            strict=strict,
            many=many,
            load_only=load_only,
            dump_only=dump_only,
            partial=partial,
        )
        kwargs = deepcopy(kwargs)
        kwargs.update(deepcopy(schema_kwargs))

        # str to set
        if isinstance(include, str):
            include = {include}

        # update kwargs from opts
        kwargs = deepcopy(kwargs)
        if isinstance(include, dict):
            opts = include.get('opts', {})
            kwargs.update(opts)

        def dump_model(model):
            """
            Tries to dump model using dumping parameters
            """
            if isinstance(model, MarshallerBase):
                _include = {}
                if isinstance(include, dict):
                    _include = include.get(key, {})
                model = model.dump(include=_include)
            return model

        dumped = self._dump(**kwargs)
        for key in include:
            if key in self.get_relationships():
                field = self.get_relationships()[key]
                val = None
                model = getattr(self, key)
                if isinstance(model, list):
                    models = [dump_model(m) for m in model]
                    val = models
                else:
                    val = dump_model(model)
                if field.dump_to:
                    key = field.dump_to
                dumped[key] = val
        return dumped

    def _dump(self, *args, **kwargs):
        """Dump helper"""

        schema = self.__class__.create_schema_instance(*args, **kwargs)
        if schema:
            return schema.dump(self).data

        # TODO: throw an exception. this is an invariant
        warnings.warn("Cannot dump! No schema attached to '{}'".format(
            self.__class__.__name__))

    def dumps(self, *schema_args, **schema_kwargs):
        """
        Dumps the model instance to a String. For parameter options, refer
        to :class:`pydent.marshaller.schema.DynamicSchema`
        """
        json_data = self.dump(*schema_args, **schema_kwargs)
        return json.dumps(json_data)

    def _fullfill(self, field):
        """
        Fullfills a relationship with a callback.

        :param field: relationship field
        :type field: Relation
        :return: model that satisfies the relationship
        :rtype: MarshallerBase
        """

        # get function
        callback = field.callback
        if not callable(callback):
            try:
                callback = getattr(self, callback)
            except AttributeError as error:
                msg = "Could not find callback \"{}\" in {} instance".format(
                    callback, self.__class__.__name__)
                error.args = tuple(list(error.args) + [msg])
                raise MarshallerCallbackNotFoundError(error)

        # get params; pass in self if param is callable
        fxn_params = self._get_callback_params(field)
        schema_model_name = field.nested
        return callback(schema_model_name, *fxn_params)

    def _get_callback_params(self, field):
        fxn_params = []
        for param in field.params:
            if callable(param):
                fxn_params.append(param(self))
            else:
                fxn_params.append(param)
        return fxn_params

    def __getstate__(self):
        """Override for pickling objects"""
        return self.dump()

    def __setstate__(self, state):
        """Override for unpickling objects"""
        self.__dict__.update(state)

    def __getattribute__(self, name):
        """Override attribute accessor to attempt to fullfill relationships.

        * If the returned attribute is None or [] and the attribute name is in
        the list of relationships, Trident will attempt to fullfill that relationship.
        * If AttributeError or TypeError is raise, return the default value"""
        res = object.__getattribute__(self, name)
        if res is None:
            relationships = object.__getattribute__(
                self, "get_relationships")()
            if name in relationships:
                try:
                    res = self.__getattr__(name)
                except AttributeError:
                    pass
                except TypeError:
                    pass
        return res

    def __getattr__(self, item):
        """
        Retrieves and fullfills relationships if available. This method runs
        only if the given attribute is not found.
        """
        relationships = object.__getattribute__(self, "get_relationships")()
        save_attr = object.__getattribute__(self, "save_attr")
        if item in relationships:
            field = relationships[item]
            ret = None
            try:
                # try to fullfill the relationships
                ret = self._fullfill(field)
            except AttributeError as error:
                # display warning
                msg = "\n"
                msg += "Could not fullfill relationship for {}(instance).{} as relation {}\nReasons:\n".format(
                    self.__class__.__name__, item, field)
                for index, message in enumerate(error.args):
                    msg += "({}) {}\n".format(index, message)
                error.args = tuple(list(error.args) + [msg])
                #warnings.warn(' '.join(error.args))
            if save_attr:
                setattr(self, item, ret)
            return ret
        raise MarshallerRelationshipError(
            "'{}' model has no attribute '{}'. Attribute was not found in list of relationships: {}".format(
                self.__class__.__name__, item,
                ', '.join(relationships.keys())))

    def _relations_to_json(self):
        """
        Dump relations to a dictionary.
        If attribute is None, return the Relation instance.
        """
        dumped = {}
        for field_name, field in self.get_relationships().items():
            val = None
            try:
                val = object.__getattribute__(self, field_name)
            except AttributeError:
                pass
            if val is None:
                val = field
            dumped[field_name] = val
        return dumped

    def _to_dict(self, *args, **kwargs):
        """
        Prints the model instance in a nice format.
        See :func:`pydent.marshaller.marshallerbase.dump`

        :param args: dump arguments
        :param kwargs: dump arguments
        :return:
        """
        dumped = self._relations_to_json()
        _dumped = self.dump(*args, **kwargs)
        if _dumped:
            for key, val in _dumped.items():
                if val is not None or key not in dumped:
                    dumped[key] = val
        return dumped

    def print(self, *args, **kwargs):
        print(pformat(self._to_dict(*args, **kwargs)))

    def __str__(self):
        return pformat(self._to_dict())

    def __repr__(self):
        return "<{}>".format(self.__class__.__name__)



"""Aquarium models

This module contains a set of classes for various various Aquarium objects.
Trident models inherit the ModelBase class and have a model schema
(generated by ``@add_schema``) that handles the JSON loading and dumping.

By default, Trident models capture ALL JSON attribute/values and sets
the attributes of the resultant object.

.. code-block:: python

    u = User.load({"id": 1, "login": "John"})
    u.login    # => "John"
    u.id       # => 1

Fields and field options are added as class variables in the model class
definition.
The various field options and their default values are listed below.

.. code-block:: python

        load_all = True     # load all data attributes defined for a model
        strict = True       # throw error during marshalling
        include = {}        # fields to include for serialization
        additional = ()     # explicit fields for serialization
        ignore = ()         # fields to filter during deserialization.
        load_only = ()      # fields to include during serialization
        dump_only = ()      # fields to include during deserialization

Trident models can also have nested relationships (for example, a Sample may
possess a single SampleType while a SampleType may possess several Samples).
These relationships can be specified in the class definition along with the
field options above, as in:

.. code-block:: python

    @add_schema
    class SampleType(Base):
        samples = fields.Nested("Sample", many=True)

In many cases, the model contains only a reference to the nested relationship,
in this case, special fields One and Many may be used to define this
relationship.
When called as an attribute, Trident will automatically use the session
connection with Aquarium to retrieve the given model.

For example, the following will define that SampleType has many Samples.
When .samples is called on a SampleType instance, Trident will use the database
to retrieve all samples that have a sample_type_id equal to the id of the
SampleType:

.. code-block:: python

    @add_schema
    class SampleType(Base):
        samples = Many("Sample",
            params={"sample_type_id": lambda self: self.id})
"""

import warnings
import shutil
import requests
import os
from pydent.base import ModelBase
from pydent.exceptions import AquariumModelError
from pydent.marshaller import add_schema, fields
from pydent.relationships import (One, Many, HasOne, HasMany,
                                  HasManyThrough, HasManyGeneric)
from pydent.utils import magiclist, filter_list
from pydent.utils.plan_validator import PlanValidator
from pydent.utils.async_requests import make_async
import json

__all__ = [
    "Account",
    "AllowableFieldType",
    "Budget",
    "Code",
    "Collection",
    "DataAssociation",
    "FieldType",
    "FieldValue",
    "Group",
    "Invoice",
    "Item",
    "Job",
    "JobAssociation",
    "Library",
    "Membership",
    "ObjectType",
    "Operation",
    "OperationType",
    "Plan",
    "PlanAssociation",
    "Sample",
    "SampleType",
    "Upload",
    "User",
    "UserBudgetAssociation",
    "Wire"
]


# Mix-ins

class HasCodeMixin:
    """Access to latest code for OperationType, Library, etc."""

    def code(self, name):
        """
        Returns the code from an OperationType or Library

        :param name: accessor name for the code.
            "source" for Library or "protocol", "precondition", "cost_model",
            or "documentation" for OperationType
        :type name: str
        :return: the code model
        :rtype: Code
        """
        if self.codes is None:
            return None
        codes = [c for c in self.codes if c.name == name]
        codes = [c for c in codes if not hasattr(
            c, "child_id") or c.child_id is None]
        if len(codes) > 0:
            return codes[-1]

    def get_code_callback(self, model_name, name):
        """
        Callback for returning code members for OperationType or Library

        :param model_name: the model_name, expected "Code"
        :type model_name: str
        :param name: accessor name for the code.
            "source" for Library or "protocol", "precondition", "cost_model",
            or "documentation" for OperationType
        :type name: str
        :return: the code model
        :rtype: Code
        """
        if model_name is not "Code":
            raise AquariumModelError(
                "Model name must be {}, not {}".format("Code", model_name))
        return self.code(name)


class FieldMixin:
    """
    Mixin for finding FieldType and FieldValue relationships
    """

    def find_field_parent(self, model_name, model_id):
        """
        Callback for finding operation_type or sample_type.
        If parent_class does not match the expected nested model name
        (OperationType or SampleType), callback will return None
        """
        if model_name == self.parent_class:
            # weird, but helps with refactoring this mixin
            fxn_name = ModelBase.find_callback.__name__
            fxn = getattr(self, fxn_name)
            return fxn(model_name, model_id)


class DataAssociatorMixin:
    """
    Mixin for handling data associations
    """

    def associate(self, key, value):
        """
        Adds a data association with the key and value to this object.
        """
        session = self.session
        aqhttp = session._AqSession__aqhttp

        data = {
            "model": {
                "model": DataAssociation.__name__,
                "record_methods": {},
                "record_getters": {}
            },
            "parent_id": self.id,
            "key": str(key),
            "object": json.dumps({str(key): value}),
            "parent_class": self.__class__.__name__,
        }

        result = aqhttp.post("json/save", json_data=data)
        return session.DataAssociation.find(result["id"])


# Models #####

@add_schema
class Account(ModelBase):
    """An Account model"""


@add_schema
class AllowableFieldType(ModelBase):
    """A AllowableFieldType model"""
    fields = dict(
        field_type=HasOne("FieldType"),
        object_type=HasOne("ObjectType"),
        sample_type=HasOne("SampleType")
    )

    def __init__(self, field_type=None, object_type=None, sample_type=None):
        self.field_type_id = None
        self.sample_type_id = None
        self.object_type_id = None
        self.field_type = self.set_model_attribute(field_type)
        self.object_type = self.set_model_attribute(object_type)
        self.sample_type = self.set_model_attribute(sample_type)
        super().__init__(**vars(self))


@add_schema
class Budget(ModelBase):
    """A Budget model"""
    fields = dict(
        user_budget_associations=HasMany("UserBudgetAssociation", "Budget")
    )


@add_schema
class Code(ModelBase):
    """A Code model"""
    fields = dict(
        user=HasOne("User"),
        operation_type=One(
            "OperationType", callback="get_parent", params=None),
        library=One("Library", callback="get_parent", params=None)
    )

    def get_parent(self, parent_class, *args):
        if parent_class != self.parent_class:
            return None
        return self.session.model_interface(
            self.parent_class).find(self.parent_id)

    def update(self):
        # since they may not always be tied to specific parent
        # controllers
        self.session.utils.update_code(self)


@add_schema
class Collection(ModelBase, DataAssociatorMixin):  # pylint: disable=too-few-public-methods
    """A Collection model"""
    fields = dict(
        object_type=HasOne("ObjectType"),
        data_associations=HasManyGeneric("DataAssociation")
    )


@add_schema
class DataAssociation(ModelBase):
    """A DataAssociation model"""
    fields = dict(
        object=fields.JSON(),
        upload=HasOne("Upload")
    )

    @property
    def value(self):
        return self.object.get(self.key, None)


@add_schema
class FieldType(ModelBase, FieldMixin):
    """A FieldType model"""
    fields = dict(
        allowable_field_types=HasMany("AllowableFieldType", "FieldType"),
        operation_type=HasOne(
            "OperationType", callback="find_field_parent", ref="parent_id"),
        sample_type=HasOne(
            "SampleType", callback="find_field_parent", ref="parent_id")
    )

    def __init__(self, name=None, ftype=None, array=None, choices=None,
                 operation_type=None, preferred_field_type_id=None,
                 preferred_operation_type_id=None, required=None, routing=None,
                 role=None, parent_class=None, parent_id=None,
                 sample_type=None, aft_stype_and_objtype=(),
                 allowable_field_types=None):
        if operation_type and sample_type:
            raise Exception()
        if operation_type is not None:
            parent_id = operation_type.id
            parent_class = "OperationType"
        if sample_type is not None:
            parent_id = sample_type.id
            parent_class = "SampleType"
        self.name = name
        self.ftype = ftype
        self.parent_id = parent_id
        self.parent_class = parent_class
        self.preferred_operation_type_id = preferred_operation_type_id
        self.preferred_field_type_id = preferred_field_type_id
        self.required = required
        self.routing = routing
        self.array = array
        self.choices = choices
        self.role = role
        self.allowable_field_types = allowable_field_types
        super().__init__(**vars(self))

        if self.allowable_field_types is None:
            if aft_stype_and_objtype is not None:
                for smple_type, obj_type in aft_stype_and_objtype:
                    self.create_allowable_field_type(smple_type, obj_type)

    @property
    def is_parameter(self):
        return self.ftype != "sample"

    def get_allowable_field_types(self):
        afts = self.allowable_field_types
        if afts is None or afts == []:
            self.allowable_field_types = None
            afts = self.allowable_field_types
        return afts

    def create_allowable_field_type(self, sample_type=None, object_type=None):
        afts = []
        if self.allowable_field_types:
            afts = self.allowable_field_types
        field_type = AllowableFieldType(
            field_type=self, sample_type=sample_type, object_type=object_type)
        afts.append(field_type)
        self.allowable_field_types = afts
        return field_type

    def initialize_field_value(self, field_value=None):
        """
        Updates or initializes a new :class:`FieldValue` from this FieldType

        :param field_value: optional FieldValue to update with this FieldType
        :type field_value: FieldValue
        :return: updated FieldValue
        :rtype: FieldValue
        """

        if not field_value:
            field_value = FieldValue(
                name=self.name, role=self.role, field_type=self)
        if self.allowable_field_types:
            field_value.allowable_field_type_id = self.allowable_field_types[0].id
            field_value.allowable_field_type = self.allowable_field_types[0]
        return field_value


@add_schema
class FieldValue(ModelBase, FieldMixin):
    """
    A FieldValue model. One of the more complex models.
    """
    fields = dict(
        # FieldValue relationships
        field_type=HasOne("FieldType"),
        allowable_field_type=HasOne("AllowableFieldType"),
        array=fields.Function(lambda fv: fv.array),
        item=HasOne("Item", ref="child_item_id"),
        sample=HasOne("Sample", ref="child_sample_id"),
        operation=HasOne(
            "Operation", callback="find_field_parent", ref="parent_id"),
        parent_sample=HasOne(
            "Sample", callback="find_field_parent", ref="parent_id"),
        wires_as_source=HasMany("Wire", ref="from_id"),
        wires_as_dest=HasMany("Wire", ref="to_id"),
        successors=HasManyThrough("Operation", "Wire"),
        sid=fields.Function(lambda fv: fv.sid, allow_none=True),
        ignore=('object_type',)
    )

    def __init__(self, name=None, role=None, parent_class=None, parent_id=None,
                 field_type=None,
                 sample=None, value=None, item=None, container=None):
        """

        :param value:
        :type value:
        :param sample:
        :type sample:
        :param container:
        :type container:
        :param item:
        :type item:
        """
        self.column = None
        self.row = None
        self.role = role
        self.id = None
        self.name = name

        self.parent_class = parent_class
        self.parent_id = parent_id
        self.operation = None  # only if parent_class == "Operation"
        self.parent_sample = None  # only if parent_class == "Sample"

        self.sample = None
        self.child_sample_id = None

        self.item = None
        self.child_item_id = None

        self.field_type = None
        self.field_type_id = None

        self.allowable_field_type = None
        self.allowable_field_type_id = None

        self.value = None
        self.sample = None
        self.item = None
        self.container = None
        # object_type is not included in the deserialization/serialization
        self.object_type = None

        if field_type is not None:
            self.set_field_type(field_type)

        if any([value, sample, item, container]):
            self._set_helper(value=value, sample=sample,
                             item=item, container=container)
        super().__init__(**vars(self))

    @property
    def array(self):
        arr = self.field_type.array
        if arr is None:
            return False
        return arr

    @property
    def sid(self):
        if self.sample is not None:
            return self.sample.identifier

    @property
    def outgoing_wires(self):
        """Alias of 'wires_as_source'"""
        return self.wires_as_source

    @property
    def incoming_wires(self):
        """Alias of 'wires_as_dest'"""
        return self.wires_as_dest

    @property
    def successors(self):
        if self.wires_as_source:
            return [x.destination for x in self.wires_as_source]
        return []

    @property
    def predecessors(self):
        if self.wires_as_dest:
            return [x.source for x in self.wires_as_dest]
        return []

    def show(self, pre=""):
        if self.sample:
            if self.child_item_id:
                item = " item: {}".format(self.child_item_id) + \
                       " in {}".format(self.item.object_type.name)
            else:
                item = ""
            print('{}{}.{}:{}{}'.format(pre, self.role,
                                        self.name, self.sample.name, item))
        elif self.value:
            print('{}{}.{}:{}'.format(pre, self.role, self.name, self.value))

    def _set_helper(self, value=None, sample=None, container=None, item=None):
        if item and container and item.object_type_id != container.id:
            raise AquariumModelError(
                "Item {} is not in container {}".format(
                    item.id, str(container)))
        if value is not None:
            self.value = value
        if item is not None:
            self.item = item
            if hasattr(item, 'id'):
                self.child_item_id = item.id
            self.object_type = item.object_type
            if not sample:
                sample = item.sample
        if sample is not None:
            self.sample = sample
            if hasattr(sample, 'id'):
                self.child_sample_id = sample.id
        if container is not None:
            self.object_type = container

    def set_value(self, value=None, sample=None, container=None, item=None):
        self._set_helper(value=value, sample=sample,
                         container=container, item=item)
        """Sets the value of a """
        if any([sample, container, item]):
            afts = self.field_type.allowable_field_types
            if self.sample is not None:
                afts = filter_list(
                    afts, sample_type_id=self.sample.sample_type_id)
            if self.object_type is not None:
                afts = filter_list(afts, object_type_id=self.object_type.id)
            if len(afts) == 0:
                available_afts = self.field_type.allowable_field_types
                msg = "No allowable field types found for {} '{}'."
                msg += " Available afts: {}"
                raise AquariumModelError(msg.format(
                    self.role, self.name, available_afts))
            if len(afts) > 1:
                msg = "More than one AllowableFieldType found that matches {}"
                warnings.warn(msg.format(
                    self.dump(only=('name', 'role', 'id'), partial=True)))
            elif len(afts) == 1:
                self.set_allowable_field_type(afts[0])
            else:
                msg = "No allowable field type found for {} '{}'"
                raise AquariumModelError(
                    msg.format(self.role, self.name))
        return self

    def set_operation(self, operation):
        self.parent_class = "Operation"
        self.parent_id = operation.id
        self.operation = operation

    def set_field_type(self, field_type):
        """
        Sets properties from a field_type
        """
        self.field_type = field_type
        self.field_type_id = field_type.id

    def set_allowable_field_type(self, allowable_field_type):
        self.allowable_field_type = allowable_field_type
        self.allowable_field_type_id = allowable_field_type.id

    def choose_item(self, first=True):
        """Set the item associated with the field value"""
        index = 0
        if not first:
            index = -1
        items = self.compatible_items()
        if items is not None and len(items) > 0:
            item = items[index]
            self.set_value(item=item)
            return item
        return None

    def compatible_items(self):
        """Find items compatible with the field value"""
        return self.session.utils.compatible_items(
            self.sample.id,
            self.allowable_field_type.object_type_id)


@add_schema
class Group(ModelBase):
    """A Group model"""
    pass


@add_schema
class Invoice(ModelBase):
    """A Invoice model"""
    pass


@add_schema
class Item(ModelBase, DataAssociatorMixin):
    """A Item model"""
    fields = dict(
        sample=HasOne("Sample"),
        object_type=HasOne("ObjectType"),
        data_associations=HasManyGeneric("DataAssociation"),
        data=fields.JSON(allow_none=True)
    )

    def __init__(self=None, sample_id=None, object_type_id=None):
        self.sample_id = sample_id
        self.object_type_id = object_type_id
        super().__init__(**vars(self))

    def make(self):
        """Makes the Item on the Aquarium server. Requires
        this Item to be connected to a session."""
        result = self.session.utils.create_items([self])
        return self.reload(result[0]["item"])

    def save(self):
        """A synonym for `make`"""
        return self.make()


@add_schema
class Job(ModelBase):
    """A Job model"""
    fields = dict(
        job_associations=HasMany("JobAssociation", "Job"),
        operations=HasManyThrough("Operation", "JobAssociation")
    )

    @property
    def uploads(self):
        http = self.session._AqSession__aqhttp
        return http.get("krill/uploads?job={}".format(self.id))['uploads']


@add_schema
class JobAssociation(ModelBase):
    """A JobAssociation model"""
    fields = dict(
        job=HasOne("Job"),
        operation=HasOne("Operation")
    )


@add_schema
class Library(ModelBase, HasCodeMixin):
    """A Library model"""
    fields = dict(
        codes=HasManyGeneric("Code"),
        source=One("Code", callback="get_code_callback", params="source")
    )


@add_schema
class Membership(ModelBase):
    fields = dict(
        user=HasOne("User"),
        group=HasOne("Group")
    )


@add_schema
class ObjectType(ModelBase):
    """A ObjectType model"""

    def save(self):
        """Saves the Object Type to the Aquarium server. Requires
        this Object Type to be connected to a session."""
        return self.reload(self.session.utils.create_object_type(self))


@add_schema
class Operation(ModelBase, DataAssociatorMixin):
    """A Operation model"""
    fields = dict(
        field_values=HasManyGeneric("FieldValue"),
        data_associations=HasManyGeneric("DataAssociation"),
        operation_type=HasOne("OperationType"),
        job_associations=HasMany("JobAssociation", "Operation"),
        jobs=HasManyThrough("Job", "JobAssociation"),
        routing=fields.Function(lambda op: op.routing, allow_none=True),
        plan_associations=HasMany("PlanAssociation", "Operation"),
        plans=HasManyThrough("Plan", "PlanAssociation")
    )

    def __init__(self, operation_type_id=None, status=None, x=0, y=0):
        self.operation_type_id = operation_type_id
        self.x = x
        self.y = y
        self.parent = 0
        self.id = None
        if status is None:
            self.status = "planning"
        self.field_values = None
        self.operation_type = None
        super().__init__(**vars(self))

    @property
    def routing(self):
        routing_dict = {}
        fvs = self.field_values
        ot = self.operation_type
        if ot is None:
            return routing_dict
        for ft in ot.field_types:
            if ft.routing is not None:
                routing_dict[ft.routing] = None
        if fvs is not None:
            for fv in self.field_values:
                ft = fv.field_type
                if ft.routing is not None:
                    routing_dict[ft.routing] = fv.sid
        return routing_dict

    def init_field_values(self):
        """
        Initialize the :class:`FieldValue` from the :class:`FieldType` of the
        parent :class:`Operation` type.
        """
        for field_type in self.operation_type.field_types:
            if field_type.array:
                self.add_to_field_value_array(field_type.name, field_type.role)
            else:
                self.set_field_value(field_type.name, field_type.role)
            # self.show()

    def field_value_array(self, name, role):
        """Returns :class:`FieldValue` array with name and role."""
        ft = self.operation_type.field_type(name, role)
        if not ft.array:
            msg = "FieldValue is not an array for the field value of operation"
            msg += " {}(id={}).{}.{}"
            raise AquariumModelError(
                msg.format(self.operation_type.name, self.id, role, name))
        return filter_list(self.field_values, name=name, role=role)

    def field_value(self, name, role):
        """
        Returns :class:`FieldValue` with name and role.
        Return None if not found.
        """
        if self.field_values:
            fvs = filter_list(self.field_values, name=name, role=role)
            if len(fvs) == 0:
                return None

            if len(fvs) == 1:
                return fvs[0]

            msg = "More than one FieldValue found for the field value"
            msg += " of operation {}(id={}).{}.{}"
            raise AquariumModelError(
                msg.format(self.operation_type.name, self.id, role, name))

    def set_field_value_array(self, name, role, values):
        """
        Sets :class:`FieldValue` array using values. Values should be a list of
        dictionaries containing sample, item, container, or values keys.
        When setting values to items/samples/containers, the
        item/sample/container must be saved.

        :param name: name of the FieldType/FieldValues being modified
        :type name: string
        :param role: role of the FieldType/FieldValue being modified
                ("input" or "output")
        :type role: string
        :param values: list of dictionary of values to set
                (e.g. [{"sample": mysample}, {"item": myitem}])
        :type values: list
        :return: the list of modified FieldValues
        :rtype: list
        """
        field_values = self.field_value_array(name, role)

        # get the field type
        field_type = self.operation_type.field_type(name, role)
        if field_type is None:
            msg = "No FieldType found for OperationType {}.{}.{}"
            raise AquariumModelError(msg.format(
                self.operation_type.name, role, name))
        if not len(field_values) == len(values):
            msg = "Cannot set_field_value_array."
            msg += " Length of values must equal length of field_values"
            raise AquariumModelError(
                msg)

        for fv, val in zip(field_values, values):
            fv.set_value(**val)
        return field_values

    def add_to_field_value_array(self, name, role, sample=None, item=None,
                                 value=None, container=None):
        """
        Creates and adds a new :class:`FieldValue`. When setting values to
        items/samples/containers, the item/sample/container must be saved.

        :param name: name of the FieldType/FieldValue
        :type name: string
        :param role: role of the FieldValue ("input" or "output")
        :type role: string
        :param sample: an existing Sample
        :type sample: Sample
        :param item: an existing Item
        :type item: Item
        :param value: a string or number value
        :type value: string|integer
        :param container: an existing ObjectType
        :type container: ObjectType
        :return: the newly created FieldValue
        :rtype: FieldValue
        """
        field_type = self.operation_type.field_type(name, role)
        fv = field_type.initialize_field_value()
        fv.set_value(sample=sample, item=item,
                     value=value, container=container)
        fv.operation = self
        if self.field_values is None:
            self.field_values = []
        self.field_values.append(fv)
        return fv

    def set_field_value(self, name, role, sample=None, item=None, value=None,
                        container=None):
        """
        Sets the a :class:`FieldValue` to a value. When setting values to
        items/samples/containers, the item/sample/container must be saved.

        :param name: name of the FieldValue/FieldType
        :type name: string
        :param role: role of the FieldValue ("input" or "output")
        :type role: string
        :param sample: an existing Sample
        :type sample: Sample
        :param item: an existing Item
        :type item: Item
        :param value: a string or number value
        :type value: string|integer
        :param container: an existing ObjectType
        :type container: ObjectType
        :return: the existing FieldValue modified
        :rtype: FieldValue
        """
        # get the field value
        field_value = self.field_value(name, role)

        # get the field type
        field_type = self.operation_type.field_type(name, role)
        if field_type is None:
            msg = "No FieldType found for OperationType {}.{}.{}"
            raise AquariumModelError(msg.format(
                self.operation_type.name, role, name))
        if field_type.array:
            msg = "FieldValue {} {} is an array. Use 'set_field_value_array'"
            raise AquariumModelError(
                msg.format(role, name))

        # initialize the field value from the field type
        if not field_value:
            field_value = field_type.initialize_field_value(field_value)
            field_value.set_operation(self)
            if self.field_values is None:
                self.field_values = []
            self.field_values.append(field_value)

        # set the value, finds allowable_field_types, etc.
        field_value.set_value(value=value, sample=sample,
                              item=item, container=container)
        return self

    @property
    def plan(self):
        return self.plans[0]

    def input(self, name):
        """Returns the input :class:`FieldValue` by name"""
        return self.field_value(name, 'input')

    def output(self, name):
        """Returns the output :class:`FieldValue` by name"""
        return self.field_value(name, 'output')

    def add_to_input_array(self, name, sample=None, item=None, value=None,
                           container=None):
        """
        Creates and adds a new input :class:`FieldValue`.
        When setting values to items/samples/containers, the
        item/sample/container must be saved.

        :param name: name of the FieldType/FieldValue
        :type name: string
        :param sample: an existing Sample
        :type sample: Sample
        :param item: an existing Item
        :type item: Item
        :param value: a string or number value
        :type value: string|integer
        :param container: an existing ObjectType
        :type container: ObjectType
        :return: the newly created FieldValue
        :rtype: FieldValue
        """
        return self.add_to_field_value_array(name, "input",
                                             sample=sample,
                                             item=item,
                                             value=value,
                                             container=container)

    def add_to_output_array(self, name, sample=None, item=None, value=None,
                            container=None):
        """
        Creates and adds a new output :class:`FieldValue`.
        When setting values to items/samples/containers, the
        item/sample/container must be saved.

        :param name: name of the FieldType/FieldValue
        :type name: string
        :param sample: an existing Sample
        :type sample: Sample
        :param item: an existing Item
        :type item: Item
        :param value: a string or number value
        :type value: string|integer
        :param container: an existing ObjectType
        :type container: ObjectType
        :return: the newly created FieldValue
        :rtype: FieldValue
        """
        return self.add_to_field_value_array(name, "output",
                                             sample=sample,
                                             item=item,
                                             value=value,
                                             container=container)

    def input_array(self, name):
        """Return a list of all input :class:`FieldValues`"""
        return self.field_value_array(name, "input")

    def output_array(self, name):
        """Return a list of all output :class:`FieldValues`"""
        return self.field_value_array(name, "output")

    def set_input(self, name, sample=None, item=None, value=None,
                  container=None):
        """
        Sets a input :class:`FieldValue` to a value. When setting values to
        items/samples/containers, the item/sample/container must be saved.

        :param name: name of the FieldValue/FieldType
        :type name: string
        :param sample: an existing Sample
        :type sample: Sample
        :param item: an existing Item
        :type item: Item
        :param value: a string or number value
        :type value: string|integer
        :param container: an existing ObjectType
        :type container: ObjectType
        :return: the existing FieldValue modified
        :rtype: FieldValue
        """
        return self.set_field_value(name, 'input', sample=sample, item=item,
                                    value=value, container=container)

    def set_output(self, name, sample=None, item=None, value=None,
                   container=None):
        """
        Sets a output :class:`FieldValue` to a value. When setting values to
        items/samples/containers, the item/sample/container must be saved.

        :param name: name of the FieldValue/FieldType
        :type name: string
        :param sample: an existing Sample
        :type sample: Sample
        :param item: an existing Item
        :type item: Item
        :param value: a string or number value
        :type value: string|integer
        :param container: an existing ObjectType
        :type container: ObjectType
        :return: the existing FieldValue modified
        :rtype: FieldValue
        """
        return self.set_field_value(name, 'output', sample=sample, item=item,
                                    value=value, container=container)

    def set_input_array(self, name, values):
        """
        Sets input :class:`FieldValue` array using values.
        Values should be a list of dictionaries containing sample, item,
        container, or values keys.
        When setting values to items/samples/containers, the
        item/sample/container must be saved.

        :param name: name of the FieldType/FieldValues being modified
        :type name: string
        :param values: list of dictionary of values to set
                (e.g. [{"sample": mysample}, {"item": myitem}])
        :type values: list
        :return: the list of modified FieldValues
        :rtype: list
        """
        return self.set_field_value_array(name, "input", values)

    def set_output_array(self, name, values):
        """
        Sets output :class:`FieldValue` array using values.
        Values should be a list of dictionaries containing sample, item,
        container, or values keys.
        When setting values to items/samples/containers, the
        item/sample/container must be saved.

        :param name: name of the FieldType/FieldValues being modified
        :type name: string
        :param values: list of dictionary of values to set
                (e.g. [{"sample": mysample}, {"item": myitem}])
        :type values: list
        :return: the list of modified FieldValues
        :rtype: list
        """
        return self.set_field_value_array(name, "output", values)

    def show(self, pre=""):
        """
        Print the operation nicely
        """
        print(pre + self.operation_type.name + " " + str(self.cost))
        for field_value in self.field_values:
            field_value.show(pre=pre + "  ")

    @property
    @magiclist
    def inputs(self):
        return [fv for fv in self.field_values if fv.role == 'input']

    @property
    @magiclist
    def outputs(self):
        return [fv for fv in self.field_values if fv.role == 'output']


@add_schema
class OperationType(ModelBase, HasCodeMixin):
    """
    A OperationType model
    """
    fields = dict(
        operations=HasMany("Operation", "OperationType"),
        field_types=Many("FieldType",
                         params=lambda self: {
                             "parent_id": self.id,
                             "parent_class": self.__class__.__name__}),
        codes=HasManyGeneric("Code"),
        protocol=One("Code", callback="get_code_callback", params="protocol"),
        cost_model=One("Code", callback="get_code_callback",
                       params="cost_model"),
        documentation=One("Code", callback="get_code_callback",
                          params="documentation"),
        precondition=One("Code", callback="get_code_callback",
                         params="precondition"),
    )

    def get_field_type(self, model_name, parent_class):
        return self.code(model_name)

    def instance(self, xpos=None, ypos=None):
        operation = Operation(operation_type_id=self.id,
                              status='planning', x=xpos, y=ypos)
        operation.connect_to_session(self.session)
        operation.operation_type = self
        operation.init_field_values()
        return operation

    def field_type(self, name, role):
        if self.field_types:
            fts = filter_list(self.field_types, role=role, name=name)
            if len(fts) > 0:
                return fts[0]

    def save(self):
        """Saves the Operation Type to the Aquarium server. Requires
        this Operation Type to be connected to a session."""
        return self.reload(self.session.utils.create_operation_type(self))


@add_schema
class Plan(ModelBase, PlanValidator, DataAssociatorMixin):
    """
    A Plan model
    """
    fields = dict(
        data_associations=HasManyGeneric("DataAssociation"),
        plan_associations=HasMany("PlanAssociation", "Plan"),
        operations=HasManyThrough("Operation", "PlanAssociation"),
        wires=Many("Wire", callback="get_wires", params=None),
        layout=fields.JSON(allow_none=True)
    )

    def __init__(self, name=None, status=None, source=None, destination=None):
        if name is None:
            name = "MyPlan"
        self.name = name
        if status is None:
            status = 'planning'
        self.id = None
        self.status = status
        self.layout = {
            "id": 0,
            "children": [],
            "documentation": "No documentation ofr this module",
            "height": 60,
            "input": [],
            "output": [],
            "name": "no_name",
            "parent_id": -1,
            "width": 160,
            "wires": []
        }
        # self.operations = []
        # self.wires = []
        self.data_associations = None
        self.plan_associations = None
        self.source = source
        self.destination = destination
        super().__init__(**vars(self))

    def add_operation(self, operation):
        """
        Adds an operation to the Plan

        :param operation: Operation to add
        :type operation: Operation
        :return: None
        :rtype: None
        """
        self.append_to_many('operations', operation)

    def add_operations(self, operations):
        """
        Adds multiple operations to the Plan

        :param operations: list of Operations
        :type operations: list
        :return: None
        :rtype: None
        """
        for operation in operations:
            self.add_operation(operation)

    def wire(self, src, dest):
        wire = Wire(source=src, destination=dest)
        self.append_to_many("wires", wire)

    def add_wires(self, pairs):
        for src, dest in pairs:
            self.wire(src, dest)

    def get_wires(self, *args):
        return self.all_wires

    # This is not being deserialized properly
    @property
    @magiclist
    def all_wires(self):
        wires = []
        if self.operations:
            for operation in self.operations:
                for field_value in operation.field_values:
                    if field_value.outgoing_wires:
                        wires += field_value.outgoing_wires
                    if field_value.incoming_wires:
                        wires += field_value.incoming_wires
        return wires

    def create(self):
        """
        Creates the Plan on the Aquarium server.

        :return: Submitted Plan (self)
        :rtype: Plan
        """
        return self.session.utils.create_plan(self)

    def submit(self, user, budget):
        """
        Submits the Plan to the Aquarium server.

        :param user: User to submit the Plan
        :type user: User
        :param budget: Budget to use for the Plan
        :type budget: Budget
        :return: JSON
        :rtype: dict
        """
        result = self.session.utils.submit_plan(self, user, budget)
        return result

    def all_data_associations(self):
        das = self.data_associations
        for operation in self.operations:
            das += operation.data_associations
            for field_value in operation.field_values:
                if field_value.item:
                    das += field_value.item.data_associations
        return das

    @classmethod
    def find(cls, session, model_id):
        """
        Override find for plans, because generic method is too minimal.
        """
        interface = cls.interface(session)
        return interface.get('plans/{}.json'.format(model_id))

    def to_save_json(self):
        """
        Returns the json representation of the plan for saving and creating
        Plans on the Aquarium server.

        :return: JSON
        :rtype: dict
        """
        json_data = self.dump(include={
            'operations': {
                'field_values': {},
            },
            'wires': {"source", "destination"}
        })
        json_data['layout'] = json.loads(json_data['layout'])

        for wire in self.wires:
            wire.dump()

        return json_data

    def estimate_cost(self):
        """
        Estimates the cost of the plan on the Aquarium server.
        This is necessary before plan submission.

        :return: cost
        :rtype: dict
        """
        return self.session.utils.estimate_plan_cost(self)

    def field_values(self):
        raise NotImplementedError()

    def show(self):
        """Print the plan nicely"""
        print(self.name + " id: " + str(self.id))
        for operation in self.operations:
            operation.show(pre="  ")
        for wire in self.wires:
            wire.show(pre="  ")

    def save(self):
        """
        Updates/patches the plan on the Aquarium server

        :return: updated Plan (self)
        :rtype: Plan
        """
        return self.session.utils.save_plan(self)

    def delete(self):
        """
        Deletes the plan on the Aquarium server

        :return: None
        :rtype: None
        """
        return self.session.utils.delete_plan(self)

    def replan(self):
        """Copies or replans the plan. Returns a plan copy"""
        return self.session.utils.replan(self.id)

    def copy(self):
        """Copies or replans the plan"""
        return self.replan()

    @staticmethod
    @make_async(1)
    def _async_download_files(data_associations, outdir=None, overwrite=True):
        Plan._download_files(data_associations, outdir=outdir, overwrite=overwrite)

    @staticmethod
    def _download_files(data_associations, outdir=None, overwrite=True):
        for da in data_associations:
            if da.upload is not None:
                da.upload.download(outdir=outdir, overwrite=overwrite)

    def download_files(self, outdir=None, overwrite=True):
        """
        Downloads all uploads associated with the plan. Downloads happen
        ansynchrounously.

        :param outdir: output directory for downloaded files
        :param overwrite: whether to overwrite files if they exist
        :return: None
        """
        return Plan._async_download_files(self.data_associations, outdir, overwrite)


@add_schema
class PlanAssociation(ModelBase):
    """A PlanAssociation model"""
    fields = dict(
        plan=HasOne("Plan"),
        operation=HasOne("Operation")
    )


@add_schema
class Sample(ModelBase):
    """A Sample model"""
    fields = dict(
        # sample relationships
        sample_type=HasOne("SampleType"),
        items=HasMany("Item", ref="sample_id"),
        field_values=HasMany("FieldValue", ref="parent_id"),
    )

    def __init__(self, name=None, project=None, sample_type_id=None):
        self.name = name
        self.project = project
        self.sample_type_id = sample_type_id
        super().__init__(**vars(self))

    @property
    def identifier(self):
        """Return the identifier used by Aquarium in autocompletes"""
        return "{}: {}".format(self.id, self.name)

    def field_value(self, name):
        for field_value in self.field_values:
            if field_value.name == name:
                return field_value
        return None

    @property
    def field_names(self):
        return [fv.name for fv in self.field_values]

    def _fv_dict(self, expected_keys):
        d = {k: None for k in expected_keys}
        fvs = self.field_values
        if fvs is not None:
            for fv in fvs:
                if fv.name in d:
                    v = None
                    if fv.sample is not None:
                        v = fv.sample
                    else:
                        v = fv.value
                    if d[fv.name] is None:
                        # set value
                        d[fv.name] = v
                    elif type(d[fv.name]) is list:
                        # append to list
                        d[fv.name].append(v)
                    else:
                        # make values an array
                        d[fv.name] = d[fv.name] + [v]
        return d

    @property
    def properties(self):
        fv_keys = [ft.name for ft in self.sample_type.field_types]
        return self._fv_dict(fv_keys)

    def save(self):
        """Saves the Sample to the Aquarium server. Requires
        this Sample to be connected to a session."""
        result = self.session.utils.create_samples([self])
        return self.reload(result['samples'][0])


@add_schema
class SampleType(ModelBase):
    """A SampleType model"""
    fields = dict(
        samples=HasMany("Sample", "SampleType"),
        field_types=Many("FieldType",
                         params=lambda self: {
                             "parent_id": self.id,
                             "parent_class": self.__class__.__name__})
    )

    def save(self):
        """Saves the Sample Type to the Aquarium server. Requires
        this Sample Type to be connected to a session."""
        return self.reload(self.session.utils.create_sample_type(self))


@add_schema
class Upload(ModelBase):
    """
    An Upload model
    """
    fields = dict(
        job=HasOne("Job")
    )

    # def _get_uploads_from_job_id(self, job_id):

    def _get_uploads_from_job(self):
        http = self.session._AqSession__aqhttp
        return http.get("krill/uploads?job={}".format(self.job_id))['uploads']

    def temp_url(self):
        uploads = self._get_uploads_from_job()
        _u_arr = [upload for upload in uploads if upload['id'] == self.id]
        return _u_arr[0]['url']

    def _download_image_from_url(self, url, outpath):
        response = requests.get(url, stream=True)
        with open(outpath, 'wb') as out_file:
            shutil.copyfileobj(response.raw, out_file)
        return response.raw

    def download(self, outdir=None, filename=None, overwrite=True):
        """
        Downloads the uploaded file

        :param outdir: path of directory of output file (default is current directory)
        :param outfile: filename of output file (defaults to upload_filename)
        :param overwrite: whether to overwrite file if it already exists
        :return: None
        """
        if outdir is None:
            outdir = '.'
        if filename is None:
            filename = "{}_{}".format(self.id, self.upload_file_name)
        filepath = os.path.join(outdir, filename)
        if not os.path.exists(filepath) or overwrite:
            self._download_image_from_url(self.temp_url(), filepath)

    @property
    def data(self):
        print(self.temp_url())
        result = requests.get(self.temp_url)
        return result.content


@add_schema
class User(ModelBase):
    """A User model"""
    fields = dict(
        groups=HasMany("Group", "User"),
        user_budget_associations=HasMany("UserBudgetAssociation", "User"),
        budgets=HasManyThrough("Budget", "UserBudgetAssociation"),
        additional=("name", "id", "login"),
        ignore=("password_digest", "remember_token", "key")
    )


@add_schema
class UserBudgetAssociation(ModelBase):
    """An association model between a User and a Budget"""
    fields = dict(
        budget=HasOne("Budget"),
        user=HasOne("User")
    )


@add_schema
class Wire(ModelBase):
    """A Wire model"""
    fields = dict(
        # load_only=False, will force dumping of FieldValues here...
        source=HasOne("FieldValue", ref="from_id", dump_to="from"),
        destination=HasOne("FieldValue", ref="to_id", dump_to="to")
    )

    def __init__(self, source=None, destination=None):
        self.source = source
        self.destination = destination
        self.active = True
        super().__init__(**vars(self))

    def to_save_json(self):
        return self.dump(include={'source', 'destination'})

    def show(self, pre=""):
        """Show the wire nicely"""
        print(pre + self.source.operation.operation_type.name +
              ":" + self.source.name +
              " --> " + self.destination.operation.operation_type.name +
              ":" + self.destination.name)



"""
Request class for making raw http requests to Aquarium

This module contains the AqHTTP class, which can make arbitrary post/put/get
requests to Aquarium and returns JSON data.

Generally, Trident users should be unable to make arbitrary requests using this
class.
Users should only access these methods indirectly through a ``Session`` or
``SessionInterface`` instance.
"""

import json
import re

import requests

from pydent.exceptions import (TridentRequestError, TridentLoginError,
                               TridentTimeoutError, TridentJSONDataIncomplete)
from pydent.utils import url_build


# TODO: Replace request history with save_attr in models?
class AqHTTP(object):
    """
    Defines a session/connection to Aquarium.
    Makes HTTP requests to Aquarium and returns JSON.

    This class should be obscured from Trident user so that users cannot make
    arbitrary requests to an Aquarium server and get sensitive information
    (e.g. User json that is returned contains api_key, password_digest, etc.)
    or make damaging posts.
    Instead, a SessionInterface should be the object that makes these requests.
    """

    TIMEOUT = 10

    def __init__(self, login, password, aquarium_url):
        """
        Initializes an aquarium session with login, password, and server.

        :param login: Aquarium login
        :type login: str
        :param aquarium_url: aquarium url to the server
        :type aquarium_url: str
        """
        self.login = login
        self.aquarium_url = aquarium_url
        self._requests_session = None
        self.timeout = self.__class__.TIMEOUT
        self._login(login, password)
        self.request_history = {}

    @staticmethod
    def create_session_json(login, password):
        """Formats login information for aquarium"""
        return {
            "session": {
                "login": login,
                "password": password
            }
        }

    @property
    def url(self):
        """An alias of aquarium_url"""
        return self.aquarium_url

    def _login(self, login, password):
        """
        Login to aquarium and saves header as a requests.Session()
        """
        session_data = self.__class__.create_session_json(login, password)
        res = None
        try:
            res = requests.post(url_build(self.aquarium_url, "sessions.json"),
                                json=session_data, timeout=self.timeout)
        except requests.exceptions.MissingSchema as error:
            raise TridentLoginError(
                "Aquairum URL {0} incorrectly formatted. {1}".format(
                    self.aquarium_url, error.args[0]))
        except requests.exceptions.ConnectTimeout:
            raise TridentTimeoutError(
                "Aquarium took too long to respond during login. Make sure "
                "the url {} is correct. Alternatively, use Session.set_timeout"
                " to increase the request timeout.".format(self.aquarium_url))
        headers = res.headers
        if 'set-cookie' not in headers:
            raise TridentLoginError(
                "Could not find proper login header for Aquarium.")
        headers = {"cookie": self.__class__.fix_remember_token(
            res.headers["set-cookie"])}
        self._requests_session = requests.Session()
        self._requests_session.headers.update(headers)

    @staticmethod
    def fix_remember_token(header):
        """ Fixes the Aquarium specific remember token """
        parts = header.split(';')
        rtok = ""
        for part in parts:
            cparts = part.split('=')
            if re.match('remember_token', cparts[0]):
                rtok = cparts[1]
        return "remember_token=" + rtok + "; " + header

    def clear_history(self):
        """Clears the request history."""
        self.request_history = {}

    @staticmethod
    def _serialize_request(url, method, body):
        return json.dumps({
            "url": url,
            "method": method,
            "body": body
        }, sort_keys=True)

    def request(self, method, path, timeout=None,
                get_from_history_ok=False, allow_none=True, **kwargs):
        """
        Performs a http request.

        :param method: request method (e.g. 'put', 'post', 'get', etc.)
        :type method: str
        :param path: url to perform the request
        :type path: str
        :param timeout: time in seconds to process request before raising
                exception
        :type timeout: int
        :param get_from_history_ok: whether its ok to return previously found
                request from history (default=False)
        :type get_from_history_ok: boolean
        :param allow_none: if False will raise error when json_data
                contains a None or null value (default: True)
        :type allow_none: boolean
        :param kwargs: additional arguments to post to request
        :type kwargs: dict
        :return: json
        :rtype: dict
        """
        if timeout is None:
            timeout = self.timeout
        if not allow_none and 'json' in kwargs:
            self._disallow_null_in_json(kwargs['json'])

        # serialize request
        url = url_build(self.aquarium_url, path)
        body = {}
        if 'json' in kwargs:
            body = kwargs['json']

        # get result from history (if ok) otherwise, make a http request;
        # save result to history
        key = self._serialize_request(url, method, body)
        result = None
        if get_from_history_ok:
            result = self.request_history.get(key, None)
        if result is None:
            result = self._requests_session.request(
                method,
                url_build(
                    self.aquarium_url, path),
                timeout=timeout,
                **kwargs)
            self.request_history[key] = result
        return self._response_to_json(result)

    @staticmethod
    def _response_to_json(result):
        """
        Turns :class:`requests.Request` instance into a json.
        Raises TridentRequestError if an error occurs.
        """
        try:
            result_json = result.json()
        except json.JSONDecodeError:
            raise TridentRequestError("Response is not JSON formatted", result)
        if result_json and 'errors' in result_json:
            raise TridentRequestError("Error response", result)
        return result_json

    @staticmethod
    def _disallow_null_in_json(json_data):
        """
        Raises :class:pydent.exceptions.TridentJSONDataIncomplete exception if
        json data being sent contains a null value
        """
        if None in json_data.values():
            raise TridentJSONDataIncomplete(
                "JSON data {} contains a null value.".format(json_data))

    def post(self, path, json_data=None, timeout=None,
             get_from_history_ok=False, allow_none=True, **kwargs):
        """
        Make a post request to the session

        :param path: url
        :type path: str
        :param json_data: json_data to post
        :type json_data: dict
        :param timeout: time in seconds to process request before raising
                exception
        :type timeout: int
        :param get_from_history_ok: whether its ok to return previously found
                request from history (default=False)
        :type get_from_history_ok: boolean
        :param allow_none: if False throw error if json_data contains a null
                or None value (default True)
        :type allow_none: boolean
        :param kwargs: additional arguments to post to request
        :type kwargs: dict
        :return: json
        :rtype: dict
        """
        return self.request("post", path, json=json_data, timeout=timeout,
                            get_from_history_ok=get_from_history_ok,
                            allow_none=allow_none, **kwargs)

    def put(self, path, json_data=None, timeout=None,
            get_from_history_ok=False, allow_none=True, **kwargs):
        """
        Make a put request to the session

        :param path: url
        :type path: str
        :param json_data: json_data to post
        :type json_data: dict
        :param timeout: time in seconds to process request before raising
                exception
        :type timeout: int
        :param get_from_history_ok: whether its ok to return previously found
                request from history (default=False)
        :type get_from_history_ok: boolean
        :param allow_none: if False throw error when json_data contains a null
                or None value (default True)
        :type allow_none: boolean
        :param kwargs: additional arguments to post to request
        :type kwargs: dict
        :return: json
        :rtype: dict
        """
        return self.request("put", path, json=json_data, timeout=timeout,
                            get_from_history_ok=get_from_history_ok,
                            allow_none=allow_none, **kwargs)

    def get(self, path, timeout=None,
            get_from_history_ok=False, allow_none=True, **kwargs):
        """
        Make a get request to the session

        :param path: url
        :type path: str
        :param timeout: time in seconds to process request before raising
                exception
        :type timeout: int
        :param get_from_history_ok: whether its ok to return previously found
                request from history (default=False)
        :type get_from_history_ok: boolean
        :param allow_none: if False throw error when json_data contains a null
                or None value (default: True)
        :type allow_none: boolean
        :param kwargs: additional arguments to post to request
        :type kwargs: dict
        :return: json
        :rtype: dict
        """
        return self.request("get", path, timeout=timeout,
                            get_from_history_ok=get_from_history_ok,
                            allow_none=allow_none, **kwargs)

    def delete(self, path, timeout=None, **kwargs):
        return self.request("delete", path, timeout=timeout, **kwargs)

    def __repr__(self):
        return "<{}({}, {})>".format(self.__class__.__name__,
                                     self.login, self.aquarium_url)

    def __str__(self):
        return self.__repr__()



"""
Utilities
"""

from functools import wraps


class MagicList(list):
    """List-like class that collects attributes and applies functions
    but acts like a list in every other regard.

    .. code-block:: python

        lst = MagicList(["string1   ", "   string2"])
        m1.strip().upper()
        #=> ["STRING1", "STRING2"]

    """

    def apply(self, fxn):
        return MagicList([fxn(x) for x in self])

    def __getitem__(self, key):
        if isinstance(key, slice):
            return MagicList(super().__getitem__(key))
        return super().__getitem__(key)

    def __getattr__(self, item):
        lst = MagicList([getattr(x, item) for x in self])
        if all([callable(x) for x in lst]):
            return MagicList(lst).call
        return lst

    def call(self, *args, **kwargs):
        return MagicList([x(*args, **kwargs) for x in self])


def magiclist(fxn):
    """
    Decorator that turns a returned value from a list to a MagicList as
    possible. Otherwise, returns original value
    """
    @wraps(fxn)
    def magiclist_wrapper(*args, **kwargs):
        ret = fxn(*args, **kwargs)
        try:
            iter(ret)
            ret = MagicList(ret)
        except TypeError:
            pass
        return ret

    return magiclist_wrapper



from marshmallow import SchemaOpts, Schema
from marshmallow import post_load, pre_load, pre_dump, post_dump
from pydent.marshaller.field_extensions import fields

# attribute to reference the model from the schema
MODEL_CLASS = "model_class"

# attribute to reference the schema from the models
MODEL_SCHEMA = "model_schema"

# class name of the field options for the schema
META = "Meta"

# class name of the field options for the models
MODEL_META = "fields"

# name of the extra fields parameter
EXTRA_FIELDS = "loaded_fields"

# attribute key to store relationships
RELATIONSHIPS = "relationships"


class DefaultFieldOptions(SchemaOpts):
    """
        Serialization/Deserialization field options class (:class:`SchemaOpts`).
        Determines how data is handled.
        There are lots of available options, how you use them will depend on
        your particular application.

        **Field options:**

        - ``load_all``: Include all attributes from the loaded data in the
                        deserialized object
        - ``ignore``: List or tuple of attributes to ignore in the deserialization
                        process
        - ``load_only``: Tuple or list of fields to exclude from serialized results.
        - ``dump_only``: Tuple or list of fields to exclude from deserialization
        - ``exclude``: Tuple or list of fields to exclude in the serialized result.
            Nested fields can be represented with dot delimiters.
        - ``include``: Dictionary of additional fields to include in the schema. It is
            usually better to define fields as class variables, but you may need to
            use this option, e.g., if your fields are Python keywords. May be an
            `OrderedDict`.
        - ``additional``: Tuple or list of fields to include *in addition* to the
            explicitly declared fields. ``additional`` and ``fields`` are
            mutually-exclusive options.
        - ``strict``: If `True`, raise errors during marshalling rather than
            storing them.

        **Additional field options (inherited from :class:`SchemaOpts`)**

        - ``fields``: Tuple or list of fields to include in the serialized result.
        - ``dateformat``: Date format for all DateTime fields that do not have their
            date format explicitly specified.
        - ``render_module``: Module to use for `loads` and `dumps`. Defaults to
            `json` from the standard library.
            Defaults to the ``json`` module in the stdlib.
        - ``ordered``: If `True`, order serialization output according to the
            order in which fields were declared. Output of `Schema.dump` will be a
            `collections.OrderedDict`.
        - ``index_errors``: If `True`, errors dictionaries will include the index
            of invalid items in a collection.
    """

    # Default field options
    load_all = True
    strict = True
    load_only = ()
    dump_only = ()
    include = {}
    additional = ()
    ignore = ()


class DynamicSchema(Schema):
    """Schema that automatically loads missing values from data. For more information about
    the baseclass Schema, take a look at the `Schema documentation
    <https://marshmallow.readthedocs.io/en/latest/api_reference.html#schema>`_ for marshmallow.
     """

    # Assign default field options
    Meta = DefaultFieldOptions
    relationships = {}

    def __init__(self, *args, only=None, exclude=(), relations=(), all_relations=False,
                 strict=None, many=False,
                 context=None, load_only=(), dump_only=(), partial=False, prefix='',
                 **kwargs):
        """
        Custom Schema for marshalling and demarshalling models


        :param tuple only: A list or tuple of fields to serialize. If `None`, all
            fields will be serialized. Nested fields can be represented with dot delimiters.
        :param tuple exclude: A list or tuple of fields to exclude from the
            serialized result. Nested fields can be represented with dot delimiters.
        :param tuple relations: A list or tuple of fields to include during
            serialization
        :param boolean all_relations: whether to dump all relationship during deserialization
        :param str prefix: Optional prefix that will be prepended to all the
            serialized field names.
        :param bool strict: If `True`, raise errors if invalid data are passed in
            instead of failing silently and storing the errors.
        :param bool many: Should be set to `True` if ``obj`` is a collection
            so that the object will be serialized to a list.
        :param dict context: Optional context passed to :class:`fields.Method` and
            :class:`fields.Function` fields.
        :param tuple load_only: A list or tuple of fields to skip during serialization
        :param tuple dump_only: A list or tuple of fields to skip during
            deserialization, read-only fields
        :param bool|tuple partial: Whether to ignore missing fields. If its value
            is an iterable, only missing fields listed in that iterable will be
            ignored.
        """

        relations, load_only, only = self.fix_schema_parameters(
            all_relations, relations, load_only, only)
        super().__init__(*args, only=only, exclude=exclude, prefix=prefix, strict=strict,
                         many=many, context=context, load_only=load_only, dump_only=dump_only,
                         partial=partial, **kwargs)
        self.dump_relations = relations

    def fix_schema_parameters(self, dump_all_relations, dump_relations, load_only, only):
        if isinstance(only, str):
            only = {only}
        if isinstance(dump_relations, str):
            dump_relations = {dump_relations}
        if isinstance(load_only, str):
            load_only = {load_only}

        # include relations in only
        if only:
            only = set(list(only) + list(dump_relations))
            only_relations = only.intersection(self.relationships.keys())
            dump_relations = set(dump_relations).union(only_relations)

        # by default, relationships are 'load_only', meaning they will not be serialized
        load_only = set(list(load_only) + list(self.relationships.keys()))

        # if there are any dump_relations keys, remove them from 'load_only' so
        # they can be serialized
        self.all_relations = dump_all_relations
        if dump_all_relations:
            dump_relations = tuple(self.relationships.keys())
        load_only = tuple(set(load_only) - set(dump_relations))

        return dump_relations, load_only, only

    @pre_load
    def add_fields(self, data):
        """Filter out ignored and load missing values"""
        # a little bit hacky, but if the data being loaded is already
        # deserialized, just reserialize it...
        if isinstance(data, getattr(self.__class__, MODEL_CLASS)):
            data = data.dump()
        self.filter_ignored(data)
        if isinstance(data, dict):
            self.load_missing(data.keys())
        return data

    def save_extra_fields(self, model_instance):
        setattr(model_instance, EXTRA_FIELDS, self.fields)

    def load_missing(self, data_keys):
        """Includes missing fields not explicitly defined in the schema"""
        meta = self.__class__.Meta
        more_keys = []
        if meta.load_all:
            for key in data_keys:
                if key not in self.fields:
                    dump_only = []
                    if hasattr(meta, "dump_only"):
                        dump_only = key in meta.dump_only
                    self.fields[key] = fields.Raw(
                        allow_none=True, dump_only=dump_only)
                    more_keys.append(key)
        setattr(meta, "additional", more_keys)

    @post_load
    def load_model(self, data):
        """
        Loads the model using the class stored in the schema when 'add_schema'
        decorator was used.
        """
        model_class = getattr(self.__class__, MODEL_CLASS)
        if model_class == data.__class__:
            return data

        # additional init parameters
        args = data.get("args", [])
        kwargs = data.get("kwargs", {})

        # instantiate model from class
        model_inst = model_class.create_from_json(*args, data=data, **kwargs)

        self.save_extra_fields(model_inst)

        # save raw JSON that created the instance
        model_inst.raw = data
        return model_inst

    def add_extra_fields(self, extra_fields):
        # copy the declared fields
        declared_fields = dict(self.declared_fields)
        self.declared_fields = declared_fields
        self.declared_fields.update(extra_fields)

    @pre_dump
    def add_extra_fields_to_dump(self, obj):
        # add extra fields if this object was loaded with extra data
        extra_fields = {}
        if hasattr(obj, EXTRA_FIELDS):
            extra_fields.update(obj.loaded_fields)

        for field_name in self.relationships.keys():
            if field_name in self.dump_relations:
                # if self._depth < self.dump_depth:
                #     extra_fields[field_name] = self.relationships[field_name]
                extra_fields[field_name] = self.relationships[field_name]
        self.add_extra_fields(extra_fields)
        return obj

    @post_dump
    def remove_ignored_from_dump(self, data):
        return self.filter_ignored(data)

    def filter_ignored(self, data):
        """Removes ignored fields from the JSON data"""
        meta = self.Meta
        ignore = []
        if hasattr(meta, "ignore"):
            ignore = meta.ignore
        for ignored in ignore:
            if ignored in data:
                del data[ignored]
        return data


def add_schema(cls):
    """
    Class Decorator that dynamically creates a :class:`DynamicSchema` and attaches it
    to a :class:`pydent.marshaller.MarshallerBase` or its subclass.

    Example usage:

    .. code-block:: python

        @add_schema
        class MyModel(MarshallerBase)
            fields = dict(
                ignore=("password",)
                # additional field options
            )

            def foo(self):
                pass
    """
    default_meta_props = {key: val for key, val in vars(
        DefaultFieldOptions).items() if not key.startswith("__")}
    model_meta_relationships = {}

    if hasattr(cls, MODEL_META):
        model_meta_props = getattr(cls, MODEL_META)
        model_meta_fields = {key: val for key, val in model_meta_props.items(
        ) if isinstance(val, fields.Field)}

        # combine found fields with "include"
        include = model_meta_props.get("include", {})
        include.update(model_meta_fields)
        model_meta_props["include"] = include

        # collect relationships
        model_meta_relationships = {
            key: val for key, val in include.items() if isinstance(val, fields.Relation)}

        # update defaults with field options found in model class definition
        default_meta_props.update(model_meta_props)

    # create the meta class
    meta = type(META, (), default_meta_props)

    # create the new schema
    new_schema = type(cls.__name__, (DynamicSchema,), {META: meta, MODEL_CLASS: cls,
                                                       RELATIONSHIPS: model_meta_relationships})
    new_schema.__name__ = new_schema.__name__ + "Schema"
    setattr(cls, MODEL_SCHEMA, new_schema)
    return cls



"""
Aquarium model baseclass

This module contains the base classes for Trident models.
Trident models load from JSON and dump to JSON.
This is accomplished by adding the ``@add_schema`` decorator to classes inherited
by the Base class.
Using ``@add_schema`` dynamically creates a model schema that handles dumping and
loading.

Features of Trident models:

    load - models can be loaded from JSON data. Hierarchical JSON is loaded
    intelligently.

.. code-block:: python

    Sample.load({"name": "MyPrimer", "sample_type": {"name": "Primer", ...} })
    # => <Sample(name="MyPrimer", sample_type=<SampleType(name="Primer")>)>

dump - models can be dumped to JSON. Dependent models and relationships can be
       dumped as well.

.. code-block:: python

    s.dump(include=("sample_type"))

relationships - models relationships are stored

.. code-block:: python

    s = Sample.load(
        {"name": "MyPrimer",
        "sample_type_id": 1}
    )

    primer_type = s.sample_type

"""

from pydent.exceptions import TridentModelNotFoundError, AquariumModelError
from pydent.marshaller import MarshallerBase
from pydent.utils import magiclist
from inflection import underscore


class ModelRegistry(type):
    """Stores a list of models that can be accessed by name."""
    models = {}

    def __init__(cls, name, bases, selfdict):
        """Saves model to the registry"""
        super().__init__(name, bases, selfdict)
        if not name == "ModelBase":
            ModelRegistry.models[name] = cls

    @staticmethod
    def get_model(model_name):
        """Gets model by model_name"""
        if model_name not in ModelRegistry.models:
            raise TridentModelNotFoundError(
                "Model \"{}\" not found in ModelRegistry.".format(model_name))
        else:
            return ModelRegistry.models[model_name]

    def __getattr__(cls, item):
        """
        Special warning for attribute errors.
        Its likely that user may have wanted to use a model interface instead of
        the Base class.
        """
        raise AttributeError("'{0}' has no attribute '{1}'. Method may be a ModelInterface method."
                             " Did you mean '<yoursession>.{0}.{1}'?"
                             .format(cls.__name__, item))


class ModelBase(MarshallerBase, metaclass=ModelRegistry):
    """
    Base class for Aquarium models.
    Subclass of :class:`pydent.marshaller.MarshallerBase`

    - creates instances from JSON using `load`
    - contains a reference to the :class:`pydent.session.aqsession.AqSession`
      instance that loaded this model
    """

    _global_record_id = 0

    def __init__(self, **model_args):
        self._session = None
        self._rid = None
        self._new_record_id()
        model_args['rid'] = self.rid
        vars(self).update(model_args)
        data = {k: v for k, v in model_args.items() if not k == '_session'}
        self._track_data(data)

    def _new_record_id(self):
        self._rid = ModelBase.new_record_id()

    @staticmethod
    def new_record_id():
        oid = ModelBase._global_record_id
        ModelBase._global_record_id += 1
        return oid

    @property
    def rid(self):
        return self._rid

    def _track_data(self, data):
        if self.model_schema:
            schema = self.model_schema()
            schema.load_missing(data.keys())
            schema.save_extra_fields(self)
            schema.validate(data)

    def append_to_many(self, name, model):
        """
        Appends a model to the many relationship

        :param name: name of the relationship or attribute
        :type name: str
        :param model: model to append
        :type model: ModelBase
        :return: None
        :rtype: None
        """
        if name in self.relationships:
            field = self.relationships[name]
            if not model.__class__.__name__ == field.model:
                raise AquariumModelError("Cannot 'append_to_many.' Model must be a '{}' but found a '{}'".format(
                    field.model, model.__class__.__name__))
            if field.many:
                val = getattr(self, name)
                if val is None:
                    val = []
                setattr(self, name, val)
                val.append(model)

    def set_model_attribute(self, model, attr="id"):
        model_name = underscore(model.__class__.__name__)
        if hasattr(model, attr):
            setattr(self, model_name + "_" + attr, getattr(model, attr))
        return model

    @classmethod
    @magiclist
    def load(cls, *args, **kwargs):
        """Create a new model instance from loaded attributes"""
        inst = super().load(*args, **kwargs)
        if isinstance(inst, list):
            for _inst in inst:
                _inst._new_record_id()
        else:
            inst._new_record_id()
        return inst

    def dump(self, *args, **kwargs):
        d = super().dump(*args, **kwargs)
        if d is not None:
            d['rid'] = self._rid
        return d

    def reload(self, data):
        """
        Reload model attributes from new data

        :param data: data to update model instance
        :type data: dict
        :return: model instance
        :rtype: ModelBase
        """
        temp_model = self.__class__.load(data=data)
        temp_model.connect_to_session(self.session)
        vars(self).update(vars(temp_model))
        return self

    @property
    def session(self):
        """The connected session instance."""
        return self._session

    def connect_to_session(self, session):
        """Connect model instance to a session. Does nothing if session already exists."""
        if self._session is None:
            self._session = session

    def _check_for_session(self):
        """Raises error if model is not connected to a session"""
        if self.session is None:
            raise AttributeError("No AqSession instance found for '{}'. Use 'connect_to_session' "
                                 "to connect this model to a session".format(self.__class__.__name__))

    def no_getter(self, *_):
        """Callback that always returns None"""
        return None

    def create_interface(self):
        return self.interface(self.session)

    @classmethod
    def interface(cls, session):
        """Creates a model interface from this class and a session"""
        return session.model_interface(cls.__name__)

    @classmethod
    def find(cls, session, model_id):
        """Finds a model instance by its model_id"""
        interface = cls.interface(session)
        return interface.find(model_id)

    @classmethod
    def where(cls, session, params):
        """Finds a list of models by some parameters"""
        if params is None:
            return None
        interface = cls.interface(session)
        return interface.where(params)

    def find_callback(self, model_name, model_id):
        """Finds a model using the model interface and model_id. Used to find
        models in model relationships."""
        self._check_for_session()
        if model_id is None:
            return None
        model = ModelRegistry.get_model(model_name)
        return model.find(self.session, model_id)

    def where_callback(self, model_name, params):
        """Finds models using a model interface and a set of parameters. Used to
        find models in model relationships."""
        self._check_for_session()
        if isinstance(params, dict):
            if len(params) == 1 and list(params.values())[0] is None:
                return None
        model = ModelRegistry.get_model(model_name)
        return model.where(self.session, params)

    def patch(self, json_data):
        """Make a patch request to self using json_data. Reload model instance with new data"""
        result = self.create_interface().patch(self.id, json_data=json_data)
        self.reload(data=result)
        return self

    def patch_with_self(self, **kwargs):
        """Update changes to this model instance to Aquarium."""
        json_data = self.dump(**kwargs)
        return self.patch(json_data=json_data)

    def __getattribute__(self, name):
        """Override getattribute to automatically connect sessions"""
        res = super().__getattribute__(name)
        if isinstance(res, list) or isinstance(res, MarshallerBase):
            relationships = object.__getattribute__(
                self, "get_relationships")()
            if name in relationships:
                session = object.__getattribute__(self, 'session')
                if isinstance(res, list):
                    [m.connect_to_session(session) for m in res]
                else:
                    res.connect_to_session(session)
        return res



"""
Session class for interacting with Trident and Aquarium

This module defines the AqSession class, which is the main entry point for
accessing the request methods.
Multiple session instances can be created using a user's Aquarium login
information.
From there, the session instance is used to make http requests
(like finding models, updating plans, etc.)

AqSession creates SessionInterfaces which can make http requests to Aquarium
but does not itself make the requests.

Interfaces are accessed by:

(1) session.<ModelName> - access the model interface
(2) session.create - access the create interface
(3) session.update - access the update interface
"""

from pydent.aqhttp import AqHTTP
from pydent.base import ModelRegistry
from pydent.interfaces import ModelInterface, UtilityInterface
from pydent.models import __all__ as allmodels
from prompt_toolkit.shortcuts import confirm
from prompt_toolkit import prompt


class AqSession(object):
    """
    Holds an AqHTTP with login information.
    Creates SessionInterfaces for models.

    .. code-block:: python

    session1 = AqSession(username, password, aquairum_url)
    session1.User.find(1)
    # <User(id=1,...)>

    """

    def __init__(self, login, password, aquarium_url, name=None):
        """

        :param login: the Aquarium login for the user
        :type login: str
        :param password: the password for the Aquarium login. This will not be
            stored anywhere. However, login headers for Aquarium will be stored
            in .__aqhttp. This may change in the future with changes
            authentication for the Aquarium API
        :type password: str
        :param aquarium_url: the http formatted Aquarium url
        :type aquarium_url: str
        :param name: (optional) name for this session
        :type name: str or None
        """
        self.name = name
        self.__aqhttp = AqHTTP(login, password, aquarium_url)
        self.__current_user = None

        # initialize model interfaces
        for model_name in allmodels:
            setattr(self, model_name,
                    ModelInterface(model_name, self.__aqhttp, self))

    def set_timeout(self, timeout_in_seconds):
        """Sets the request timeout."""
        self.__aqhttp.timeout = timeout_in_seconds

    @property
    def url(self):
        """Returns the aquarium_url for this session"""
        return self.__aqhttp.aquarium_url

    @property
    def login(self):
        """
        Logs into aquarium, generating the necessary headers to perform requests
        to Aquarium
        """
        return self.__aqhttp.login

    @classmethod
    def interactive(cls):
        """
        Login using prompts and a hidden password (********)
        """
        confirm_register = False
        password = None
        username = None
        url = None
        while not confirm_register:
            password = None
            confirm_password = None
            name = prompt('enter name for session: ')
            username = prompt('enter username: ')
            url = prompt('enter url: ')
            msg = ''
            while confirm_password is None or password != confirm_password:
                if msg:
                    print(msg)
                password = prompt('enter password: ', is_password=True)
                confirm_password = prompt(
                    'confirm password: ', is_password=True)
                msg = "passwords did not match!"
            confirm_register = confirm(
                'Confirm registration for {}@{}? (y/n): '.format(username, url))
            print()
        print("username {} registered".format(username))
        return cls(login=username, password=password, aquarium_url=url, name=name)

    @property
    def current_user(self):
        """
        Returns the current User associated with this session. Returns None
        if no user is found (as in cases where the Aquarium connection is down).
        """
        if self.__current_user is None:
            user = self.User.where({"login": self.__aqhttp.login})
            if not user:
                return None
            self.__current_user = self.User.where(
                {"login": self.__aqhttp.login})[0]
        return self.__current_user

    def logged_in(self):
        """
        Returns whether the user is logged in. If the session
        is able to return the User model instance using the
        session's login credentials, the user is considered
        to be logged in.

        :return: whether user is currently logged in
        :rtype: boolean
        """
        if self.current_user is None:
            return False
        return True

    @property
    def models(self):
        """Returns list of all models available"""
        return list(ModelRegistry.models.keys())

    def model_interface(self, model_name):
        """Returns model interface by name"""
        return ModelInterface(model_name, self.__aqhttp, self)

    @property
    def utils(self):
        """Instantiates a utility interface"""
        return UtilityInterface(self.__aqhttp, self)

    def __repr__(self):
        return "<{}(name={}, AqHTTP={}))>".format(self.__class__.__name__,
                                                  self.name, self.__aqhttp)



"""
Trident exceptions
"""

from requests.exceptions import RequestException, ConnectTimeout


class TridentRequestError(RequestException):
    """There was an ambiguous exception that occured handling your request."""
    def __init__(self, message, response):
        self.message = message
        self.response = response


class TridentJSONDataIncomplete(RequestException):
    """JSON data contains a null value and may be incomplete."""


class TridentLoginError(RequestException):
    """Trident is not properly connected to the server.
     Verify login credentials are correct."""


class TridentTimeoutError(ConnectTimeout):
    """Trident took too long to respond"""


class TridentModelNotFoundError(AttributeError):
    """Trident could not find model in list of models."""


class AquariumModelError(Exception):
    """An error occured with this Aquarium model"""



"""

.. module:: pydent

Submodules
==========

.. autosummary::
    :toctree: _autosummary

    aqhttp
    aqsession
    base
    exceptions
    interfaces
    models
    relationships

"""

from .__version__ import __description__, __author__, __version__, __url__, __title__
from .aqsession import AqSession
from .base import ModelBase, ModelRegistry
from .utils import pprint



"""Serialization/deserialization

==========
marshaller
==========

Submodules
==========

.. autosummary::
    :toctree: _autosummary

    marshallerbase
    field_extensions
    schema
    exceptions

"""

from pydent.marshaller.marshallerbase import MarshallerBase
from pydent.marshaller.field_extensions import fields, Relation, JSON
from pydent.marshaller.schema import add_schema



"""
Marsaller exceptions
"""


class MarshallerCallbackNotFoundError(Exception):
    """Could not find callback for the model."""


class MarshallerRelationshipError(AttributeError):
    """Could not fullfill relationship."""



import asyncio
from functools import wraps
from tqdm import tqdm
import itertools


def chunkify(list, n):
    """Breaks a list into n chunks. Last chunk may not be equal in size to other chunks"""
    l = list
    return [l[i:i + n] for i in range(0, len(l), n)]


async def exec_async_fxn(fxn, args, extra_args=list(), chunk_size=1, desc='', progress_bar=True):
    """Executes an asynchronous function"""
    # run asynchronously
    loop = asyncio.get_event_loop()
    futures = [
        loop.run_in_executor(None, fxn, arg, *extra_args)
        for arg in args
    ]

    # collect results
    results = []

    iterator = asyncio.as_completed(futures)
    if progress_bar:
        iterator = tqdm(iterator, desc=desc, total=len(futures), unit_scale=chunk_size)

    for f in iterator:
        results.append(await f)
    return results


def asyncfunc(fxn, arg_chunks, args=list(), chunk_size=1, progress_bar=True, desc=None):
    """
    Runs a function asynchronously.

    :param fxn: function to run asynchronously
    :type fxn: function or lambda
    :param arg_chunks: arguments to apply to the function; suggested to divide list into chunks
    :type arg_chunks: list
    :return: result
    :rtype: list
    """
    # finish loop
    loop = asyncio.get_event_loop()
    if desc is None:
        desc = desc
    results = loop.run_until_complete(exec_async_fxn(fxn, arg_chunks, extra_args=args, desc=desc, progress_bar=progress_bar, chunk_size=chunk_size))
#     loop.close()
    return results


def make_async(chunk_size, progress_bar=True):
    def dec(fxn):
        @wraps(fxn)
        def wrapper(data, *args):
            chunks = chunkify(data, chunk_size)

            desc = "Running \"{}\" [size: {}, num: {}]: ".format(fxn.__name__, chunk_size, len(chunks))
            results = asyncfunc(fxn, chunks, args=args, progress_bar=progress_bar, desc=desc)
            return list(itertools.chain(*results))
        return wrapper
    return dec


import os
import re
import sys
from distutils.core import setup
from setuptools.command.install import install


tests_require = [
    'pytest',
    'pytest-cov'
]

install_requires = [
    'inflection',
    'marshmallow==2.15.1',
    'requests',
    'prompt-toolkit==2.0.3',
    'tqdm==4.23.4'
]

def parse_version_file():
    here = os.path.abspath(os.path.dirname(__file__))
    ver_dict = {}
    with open(os.path.join(here, 'pydent', '__version__.py'), 'r') as f:
        for line in f.readlines():
            m = re.match('__(\w+)__\s*=\s*(.+)', line)
            if m:
                key = m.group(1)
                val = m.group(2)
                val = re.sub("[\'\"]", "", val)
                ver_dict[key] = val
    return ver_dict

def readme():
    """print long description"""
    with open('README.rst') as f:
        return f.read()


class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version"""
    description = 'verify that the git tag matches our version'

    def run(self):
        tag = os.getenv('CIRCLE_TAG')

        if tag != ver['version']:
            info = "Git tag: {0} does not match the version of this app: {1}".format(
                tag, ver['version']
            )
            sys.exit(info)


ver = parse_version_file()

# setup
setup(
        title=ver['title'],
        name='pydent',
        version=ver['version'],
        packages=["pydent", "pydent.marshaller", "pydent.utils"],
        long_description=readme(),
        url=ver['url'],
        license='',
        author=ver['author'],
        author_email=ver['author_email'],
        keywords='aquarium api',
        description=ver['description'],
        install_requires=install_requires,
        python_requires='>=3.4',
        tests_require=tests_require,
        classifiers=[],
        cmdclass={
            'verify': VerifyVersionCommand,
        }
)



"""
Module containing custom :class:`marshmallow.fields`.
"""

import json
from marshmallow import fields, ValidationError


class JSON(fields.Field):
    """A custom JSON field"""

    def _serialize(self, value, attr, obj):
        if value is None:
            return ''
        try:
            return json.dumps(value)
        except TypeError as error:
            raise ValidationError(error)

    def _deserialize(self, value, attr, data):
        if value is None:
            return ''
        if isinstance(value, dict):
            return value
        try:
            return json.loads(value)
        except json.decoder.JSONDecodeError as error:
            raise ValidationError(error)

    default_error_message = {
        'invalid': 'field was unable to be parsed as a JSON'
    }


class Relation(fields.Nested):
    """
    Defines a nested relationship with another model.

    Uses "callback" with "params" to find models.
    Callback is applied to the model that is fullfilling this relation.
    Params may be lambdas of the form "lambda self: <do something with self>"
    which passes in the model instance.
    """

    def __init__(self, model, callback, params, *args, allow_none=True, **kwargs):
        """
        Relation initializer.

        :param model: target model
        :type model: basestring
        :param args: positional parameters
        :type args:
        :param callback: function to use in Base to find model
        :type callback: basestring or callable
        :param params: tuple or list of variables (or callables) to use to
                       search for the model. If param is a callable, the model
                       instance will be passed in.
        :type params: tuple or list
        :param kwargs: rest of the parameters
        :type kwargs:
        """
        super().__init__(model, *args, allow_none=allow_none, **kwargs)
        self.callback = callback
        # force params to be an iterable
        if not isinstance(params, (list, tuple)):
            params = (params,)
        self.params = params

    @property
    def model(self):
        return self.nested

    def _serialize(self, nested_obj, attr, obj):
        dumped = None

        def dump(nobj):
            if hasattr(nobj, 'dump'):
                nobj = nobj.dump()
            return nobj

        if nested_obj:
            if isinstance(nested_obj, list):
                dumped = []
                for model in nested_obj:
                    xdumped = None
                    if model is not None:
                        xdumped = dump(model)
                    dumped.append(xdumped)
            else:
                dumped = dump(nested_obj)
        return dumped

    def __repr__(self):
        return "<{} (model={}, callback={}, params={})>".format(
            self.__class__.__name__,
            self.nested, self.callback, self.params)


fields.Relation = Relation
fields.JSON = JSON



"""Validate Plans"""


class PlanValidator(object):
    """
    Used to validate whether a plan is ready to submit.
    """

    def __init__(self, model, data):
        """
        Creates a PlanValidator for the ?
        """
        self.verbose = False
        self.messages = []
        self.name = ""
        self.operations = []
        self.wires = []
        super(PlanValidator, self).__init__(model, data)

    def log(self, msg):
        """
        Log a message.

        :param msg: the message to be logged
        :type msg: String
        """
        if self.verbose:
            print(msg)
        self.messages.append(msg)

    def warning(self, msg):
        """
        Log a warning.

        :param msg: the warning message to be logged
        :type msg: String
        """
        self.log("WARNING: " + msg)

    def clear_messages(self):
        """
        Clear messages. 
        """
        self.messages = []

    def validate(self, verbose=False):
        """
        Check if the plan is valid.
        """
        self.verbose = verbose
        self.clear_messages()
        self.mark_leaves()

        valid = self.check_fields() and \
            self.check_io() and \
            self.check_wires()

        self.log("Plan '" + self.name + "' is " +
                 ("valid" if valid else "not valid."))

        return valid

    def check_fields(self):
        """Check that all fields are defined"""
        valid = True
        for operation in self.operations:
            for field_type in operation.operation_type.field_types:
                if field_type.array:
                    if len(operation.field_value_array(field_type.name, field_type.role)) == 0:
                        self.warning(field_type.role + " " + field_type.name +
                                     "' of '" + field_type.operation_type.name +
                                     "' cannot be an empty array.")
                        valid = False
                else:
                    if not operation.field_value(field_type.name, field_type.role):
                        self.warning(field_type.role + " " + field_type.name +
                                     "' of '" + field_type.operation_type.name +
                                     "' has not been assigned.")
                        valid = False
        return valid

    def mark_leaves(self):
        """mark all inputs as either a leaf or not"""
        for operation in self.operations:
            for field_value in operation.outputs:
                field_value.leaf = False
            for field_value in operation.inputs:
                field_value.leaf = True

        for wire in self.all_wires:
            wire.destination.leaf = False

    def check_io(self):
        """check all field values for samples, items, and values"""
        valid = True
        for operation in self.operations:
            for field_value in operation.field_values:
                if field_value.field_type.is_parameter:
                    if not field_value.value:
                        self.warning("Parameter '" + field_value.name +
                                     "' of '" + field_value.operation.operation_type.name +
                                     "' has not been assigned a value.")
                        valid = False
                elif field_value.leaf:
                    if not field_value.child_item_id:
                        self.warning("Leaf '" + field_value.name +
                                     "' of '" + field_value.operation.operation_type.name +
                                     "' has not been assigned an item")
                        valid = False
                else:
                    if not field_value.child_sample_id and \
                       field_value.allowable_field_type.sample_type:
                        self.warning(field_value.role +
                                     " '" + field_value.name +
                                     "' of '" + field_value.operation.operation_type.name +
                                     "' has not been assigned a sample")
                        valid = False
        return valid

    def check_wires(self):
        """make sure all wires have have consistent endpoints"""
        valid = True
        for wire in self.all_wires:
            stid1 = wire.source.allowable_field_type.sample_type_id
            stid2 = wire.destination.allowable_field_type.sample_type_id
            otid1 = wire.source.allowable_field_type.object_type_id
            otid2 = wire.destination.allowable_field_type.object_type_id
            if stid1 != stid2 or otid1 != otid2:
                valid = False
        return valid
